#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The graphical interface for submitting SEAMM jobs.

A job in SEAMM is composed of a flowchart and any other files that the
flowchart requires. This module provides the TkJobHandler class, which
provides a use interface and the machinery to gather the necessary
files and submit the job to a dashboard.
"""

import configparser
import logging
from pathlib import Path
import pkg_resources
import shlex
import tkinter as tk
from tkinter import messagebox
from tkinter import simpledialog
import tkinter.ttk as ttk

import Pmw

from .dashboard_handler import DashboardHandler
from seamm_dashboard_client import DashboardConnectionError, DashboardLoginError
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkJobHandler(object):
    def __init__(self, root=None):
        """Setup the Job Handler object.

        Parameters
        ----------
        root : Tk window
            The root Tk window.
        """
        self._root = root
        self.config = configparser.ConfigParser()
        self.dialog = None

        self._dashboard_handler = None

        self._flowchart = None
        self._widgets = {}
        self._variable_value = {}
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))

        s = ttk.Style()
        s.configure("Border.TLabel", relief="ridge", anchor=tk.W, padding=5)

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets."""
        return self._widgets[key]

    def __setitem__(self, key, value):
        """Allow [key] access to set a widgets."""
        self._widgets[key] = value

    def __delitem__(self, key):
        """Allow deletion of widgets."""
        try:
            if key in self._widgets:
                self._widgets[key].destroy()
        except Exception:
            pass
        del self._widgets[key]

    def __iter__(self):
        """Allow iteration over the widgets"""
        return iter(self._widgets)

    def __len__(self):
        """Provide the nmber of widgets, for e.g. len() command."""
        return len(self._widgets)

    @property
    def current_dashboard(self):
        "The current dashboard, from dashboard_handler"
        return self.dashboard_handler.current_dashboard

    @current_dashboard.setter
    def current_dashboard(self, dashboard):
        self.current_dashboard.current_dashboard = dashboard

    @property
    def dashboard_handler(self):
        "The connection to the dashboards."
        if self._dashboard_handler is None:
            self._dashboard_handler = DashboardHandler()
        return self._dashboard_handler

    def add_dashboard_cb(self):
        """Post a dialog for adding a dashboard to the list."""
        dialog = Pmw.Dialog(
            self._root,
            buttons=("OK", "Cancel"),
            master=self._root,
            title="Add Dashboard to list",
            command=self.handle_add_dialog,
        )
        dialog.withdraw()
        w = self["add"] = {"dialog": dialog}

        d = dialog.interior()
        name = sw.LabeledEntry(d, labeltext="Name", width=50)
        url = sw.LabeledEntry(d, labeltext="URL")
        protocol = sw.LabeledCombobox(
            d, labeltext="Protocol", values=["http", "sshtunnel"]
        )
        protocol.set("http")

        w["name"] = name
        w["url"] = url
        w["protocol"] = protocol

        name.grid(row=0, column=0, sticky=tk.EW)
        url.grid(row=1, column=0, sticky=tk.EW)
        protocol.grid(row=2, column=0, sticky=tk.W)

        sw.align_labels([name, url, protocol])

        dialog.activate(geometry="centerscreenfirst")

    def ask_for_credentials(self, dashboard, user=None, password=None):
        """Prompt the user for the login for the dashboard

        Parameters
        ----------
        dashboard : str
            The name of the dashboard.
        user : str
            The username for that dashboard.
        password : str
            The password for the user.

        Returns
        -------
        (str, str)
            A tuple with the username and password.
        """
        dialog = Pmw.Dialog(
            self._root,
            buttons=("OK", "Cancel"),
            master=self._root,
            title=f"Log-in for {dashboard}",
        )
        dialog.withdraw()

        d = dialog.interior()
        w_user = sw.LabeledEntry(d, labeltext="Username:", width=50)
        w_password = sw.LabeledEntry(d, labeltext="Password:", show="*")

        w_user.grid(row=0, column=0, sticky=tk.EW)
        w_password.grid(row=1, column=0, sticky=tk.EW)

        sw.align_labels([w_user, w_password], sticky=tk.E)

        result = dialog.activate(geometry="centerscreenfirst")

        if result == "OK":
            user = w_user.get()
            password = w_password.get()

        dialog.destroy()

        return user, password

    def check_status_cb(self):
        """Helper for checking the status of a dashboard."""
        w = self["edit dashboard"]
        dashboard = w["dashboard"]
        status = self.dashboard_handler.get_dashboard(dashboard).status()
        w["status"].set(status)

    def create_submit_dialog(self, title="", description=""):
        """Create the dialog for submitting a job.

        Parameters
        ----------
        flowchart : seamm.Flowchart
            The flowchart object
        """
        logger.debug("Creating submit dialog")
        self.dialog = Pmw.Dialog(
            self._root,
            buttons=("OK", "Cancel"),
            master=self._root,
            title="Submit job to SEAMM",
            command=self.handle_dialog,
        )
        self.dialog.withdraw()

        d = self.dialog.interior()

        # Dashboard
        dashboards = self.dashboard_handler.dashboards
        self["dashboard"] = sw.LabeledCombobox(
            d, labeltext="Dashboard:", values=dashboards
        )
        self["dashboard"].combobox.bind("<<ComboboxSelected>>", self.dashboard_cb)

        self["add"] = ttk.Button(
            d, text="add dashboard...", command=self.add_dashboard_cb
        )

        # User and project
        self["project"] = sw.LabeledCombobox(d, labeltext="Project:", state="readonly")
        self["project"].bind("<<ComboboxSelected>>", self.project_cb)

        # Title
        self["title"] = sw.LabeledEntry(d, labeltext="Title:", width=100)
        self["title"].set(title)

        # Description
        self["description label"] = ttk.Label(d, text="Description:")
        frame = sw.ScrolledFrame(
            d, scroll_vertically=True, borderwidth=2, relief=tk.SUNKEN
        )
        f = frame.interior()
        self["description"] = tk.Text(f)
        self["description"].grid(sticky=tk.EW)
        self["description"].insert("1.0", description)

        f.rowconfigure(0, weight=1)
        f.columnconfigure(0, weight=1)

        # Reset and clear buttons
        rf = self["reset frame"] = tk.Frame(d)
        self["reset title"] = ttk.Button(
            rf, text="reset title", command=self.reset_title
        )
        self["clear title"] = ttk.Button(
            rf, text="clear title", command=self.clear_title
        )
        self["reset description"] = ttk.Button(
            rf, text="reset description", command=self.reset_description
        )
        self["clear description"] = ttk.Button(
            rf, text="clear description", command=self.clear_description
        )
        self["reset title"].grid(row=0, column=0)
        self["clear title"].grid(row=0, column=1)
        self["reset description"].grid(row=0, column=2)
        self["clear description"].grid(row=0, column=3)

        # Space for any parameters
        self["parameters label"] = ttk.Label(d, text="Parameters:")
        self["parameters"] = sw.ScrolledColumns(
            d,
            columns=[
                "Name",
                "Value",
                "",
                "Description",
            ],
        )

        # Set up the dashboard and projects if needed
        if len(dashboards) > 0:
            self["dashboard"].set(self.current_dashboard.name)
            self.dashboard_cb()

        # Grid the widgets into rows and columns
        self["dashboard"].grid(row=0, column=0, sticky=tk.EW)
        self["add"].grid(row=0, column=1, sticky=tk.W)
        self["project"].grid(row=1, column=0, sticky=tk.EW)
        self["title"].grid(row=2, column=0, columnspan=2, sticky=tk.W)
        self["description label"].grid(row=3, column=0, columnspan=2, sticky=tk.W)
        frame.grid(row=4, column=0, columnspan=2, sticky=tk.NSEW)
        self["reset frame"].grid(row=5, column=0, columnspan=2)
        self["parameters label"].grid(row=6, column=0, columnspan=2, sticky=tk.W)
        self["parameters"].grid(row=7, column=0, columnspan=2, sticky=tk.NSEW)

        sw.align_labels([self["dashboard"], self["project"], self["title"]])

        d.rowconfigure(4, weight=1)
        d.rowconfigure(7, weight=1)
        d.columnconfigure(1, weight=1)

    def dashboard_cb(self, event=None):
        """The selected dashboard has been changed"""
        dashboard = self["dashboard"].get()

        # Ensure that user has account on dashboard
        user, passwd = self.dashboard_handler.get_credentials(
            dashboard, ask=self.ask_for_credentials
        )
        if user is None or passwd is None:
            # Unable to log in, so go back
            if self.current_dashboard is not None:
                dashboard = self.current_dashboard.name
                self["dashboard"].set(dashboard)
                return

        try:
            projects = self.dashboard_handler.get_dashboard(dashboard).list_projects()
        except DashboardLoginError as e:
            msg = e.args[1]
            messagebox.showwarning("Unable to login", msg)
            return
        except DashboardConnectionError as e:
            msg = e.args[1]
            messagebox.showwarning("Unable to reach the dashboard", msg)
            return

        if len(projects) == 0:
            if self.current_dashboard is not None:
                self["dashboard"].set(self.current_dashboard.name)
            return

        # All OK, changed the widgets
        projects.append("-- Create new project --")
        self["project"].combobox.config({"value": projects})
        if len(projects) > 0:
            self["project"].set(projects[0])
        else:
            self["project"].set("")
        self.current_dashboard = dashboard

    def display_dashboards(self):
        """Display a list of all the dashboards with their status.

        Allow users to edit, remove and add dashboards.
        """
        # statuses = self.get_all_status(master=self._root)

        dialog = Pmw.Dialog(
            self._root,
            buttons=("OK", "Edit", "Remove", "Cancel"),
            master=self._root,
            title="Dashboards",
            command=self.handle_dashboard_dialog,
        )
        dialog.withdraw()
        w = self["display"] = {"dialog": dialog}

        d = dialog.interior()
        w["table"] = sw.ScrolledColumns(
            d,
            columns=[
                "Select",
                "Dashboard",
                "URL",
                "Status",
            ],
            header_style="Border.TLabel",
        )
        w["table"].grid(row=0, column=0, sticky=tk.NSEW)
        d.columnconfigure(0, weight=1)
        d.rowconfigure(0, weight=1)

        # Button to update all status
        w["update all"] = ttk.Button(
            d, text="Update Status of All Dashboards", command=self.fill_statuses
        )
        w["update all"].grid()

        # Fill in the dashboards
        table = w["table"]
        f = table.interior()
        row = 0
        w["selected"] = tk.StringVar()
        if self.current_dashboard is not None:
            w["selected"].set(self.current_dashboard.name)
        for dashboard in self.dashboard_handler.dashboards:
            table[row, 0] = ttk.Radiobutton(
                f,
                variable=w["selected"],
                value=dashboard,
            )
            table[row, 1] = ttk.Label(f, text=dashboard, style="Border.TLabel")
            table[row, 2] = ttk.Label(
                f, text=self.config[dashboard]["url"], style="Border.TLabel"
            )
            state = self.config.get(dashboard, "state", fallback="active")
            if state == "active":
                table[row, 3] = ttk.Label(f, width=16, style="Border.TLabel")
            else:
                table[row, 3] = ttk.Label(
                    f, text=state, width=16, style="Border.TLabel"
                )
            table[row, 0].grid(sticky=tk.EW)
            table[row, 1].grid(sticky=tk.EW)
            table[row, 2].grid(sticky=tk.EW)
            table[row, 3].grid(sticky=tk.EW)
            row += 1

        self.fit_dialog(dialog)

        dialog.activate(geometry="centerscreenfirst")

        dialog.destroy()
        del self["display"]

    def edit_cb(self, dashboard):
        """Edit the information for a dashboard."""
        table = self["display"]["table"]
        for trow in range(table.nrows):
            if table[trow, 1].cget("text") == dashboard:
                break

        dialog = Pmw.Dialog(
            self._root,
            buttons=("OK", "Cancel"),
            title="Edit " + dashboard.title(),
        )
        dialog.withdraw()
        w = self["edit dashboard"] = {"dialog": dialog, "dashboard": dashboard}

        d = dialog.interior()
        name = sw.LabeledEntry(d, labeltext="Name:", width=50)
        name.set(table[trow, 1].cget("text"))
        url = sw.LabeledEntry(d, labeltext="URL:", width=50)
        url.set(table[trow, 2].cget("text"))
        state = sw.LabeledCombobox(d, labeltext="State:", values=["active", "inactive"])
        current_status = table[trow, 3].cget("text")
        if current_status == "inactive":
            state.set("inactive")
        else:
            state.set("active")
        status = sw.LabeledEntry(d, labeltext="Status:", width=16)
        status.set(current_status)
        check_status = ttk.Button(d, text="Check Status", command=self.check_status_cb)
        w["status"] = status

        name.grid(row=0, columnspan=2, sticky=tk.EW)
        url.grid(row=1, columnspan=2, sticky=tk.EW)
        state.grid(row=2, columnspan=2, sticky=tk.EW)
        status.grid(row=3, sticky=tk.EW)
        check_status.grid(row=3, column=1)

        widgets = [name, url, state, status]

        # If there are any other items in the config file, put them in
        row = 3
        for key, value in self.config.items(dashboard):
            if key not in ("url", "state", "status"):
                row += 1
                w[key] = sw.LabeledEntry(d, labeltext=key + ":")
                w[key].set(value)
                w[key].grid(row=row, columnspan=2, sticky=tk.EW)
                widgets.append(w[key])

        sw.align_labels(widgets)

        result = dialog.activate(geometry="centerscreenfirst")

        if result == "OK":
            if name.get() != dashboard:
                self.dashboard_handler.rename_dashboard(dashboard, name.get())
                # Changed the name. Move the section in the config file
                dashboard = name.get()
                table[trow, 1].configure(text=dashboard)
                table[trow, 0].configure(value=dashboard)
                self["display"]["selected"].set(dashboard)
                self.current_dashboard = dashboard

            dboard = self.dashboard_handler.get_dashboard(dashboard)

            table[trow, 2].configure(text=url.get())
            dboard["url"] = url.get()
            if state.get() == "active":
                if status.get() == "inactive":
                    table[trow, 3].configure(text="")
                else:
                    table[trow, 3].configure(text=status.get())
            else:
                table[trow, 3].configure(text=state.get())
            dboard["state"] = state.get()

            for key, value in self.config.items(dashboard):
                if key not in ("url", "state", "status"):
                    dboard[key] = w[key].get()

            self.dashboard_handler.update(dashboard)

    def file_cb(self, table, row, name, data):
        """Method to handle parameters with files

        Parameters
        ----------
        table : sw.ScrolledColumns
            The widget displaying the table of parameters.
        row : int
            The row of the table.
        name : str
            The name of the parameter.
        data : dict(str, str)
            The definition of the parameter.
        """
        multiple = data["nargs"] != "a single value"

        filetypes = [
            ("MOL", "*.mol"),
            ("MOL", "*.mol2"),
            ("SDF", "*.sdf"),
            ("XYZ", "*.xyz"),
            ("CIF", "*.cif"),
            ("MMCIF", "*.mmcif"),
            ("All files", "*"),
        ]
        filename = tk.filedialog.askopenfilename(filetypes=filetypes, multiple=multiple)
        if filename == "":
            return
        w = table[row, 1]
        if multiple:
            current = shlex.split(w.get())
            for name in filename:
                if name not in current:
                    current.append(name)
            w.delete(0, tk.END)
            w.insert(0, " " + shlex.join(current))
        else:
            w.delete(0, tk.END)
            w.insert(0, filename)

    def fill_statuses(self):
        w = self["display"]
        dialog = w["dialog"]
        table = w["table"]

        dialog.configure(title="Dashboards -- Updating Status")
        progress = ttk.Progressbar(
            dialog.interior(),
            orient=tk.HORIZONTAL,
            maximum=table.nrows + 1,
            mode="determinate",
            value=1,
        )
        progress.grid(sticky=tk.EW)

        for row in range(table.nrows):
            current_status = table[row, 3].cget("text")
            if current_status != "inactive":
                table[row, 3].configure(text="...")
                table.update()
                dashboard = table[row, 1].cget("text")
                status = self.dashboard_handler.get_dashboard(dashboard).status()
                table[row, 3].configure(text=status)
            progress.step()
            table.update()
        progress.destroy()
        dialog.configure(title="Dashboards")

    def fit_dialog(self, dialog):
        """Resize and fit the dialog to the current contents and the
        constraint of the window.
        """
        logger.debug("Entering fit_dialog")

        widget = self["display"]["table"].interior()
        logger.debug("  widget = {}".format(widget))
        widget.update_idletasks()

        frame = dialog.interior()
        frame.update_idletasks()
        width = frame.winfo_width()
        height = frame.winfo_height()
        sw = frame.winfo_screenwidth()
        sh = frame.winfo_screenheight()

        logger.debug(
            "  frame wxh = {} x {}, screen = {} x {}".format(width, height, sw, sh)
        )

        mw = frame.winfo_reqwidth()
        mh = frame.winfo_reqheight()
        logger.debug("  frame requested = {} x {}".format(mw, mh))

        # Need to handle scrolledtable using its inside frame
        ww = widget.winfo_width()
        hh = widget.winfo_height()
        w = widget.winfo_reqwidth()
        h = widget.winfo_reqheight()
        logger.debug("  table wxh = {} x {}, requested = {} x {}".format(ww, hh, w, h))
        h += 20  # Add a bit of space...
        if w > mw:
            mw = w
        if h > mh:
            mh = h
        if ww > width:
            width = ww
        if hh > height:
            height = hh

        if width < mw:
            width = mw
        width += 70
        if width > 0.9 * sw:
            width = int(0.9 * sw)
        if height < mh:
            height = mh
        height += 70
        if height > 0.9 * sh:
            height = int(0.9 * sh)

        dialog.geometry("{}x{}".format(width, height))

    def get_all_status(self, show_progress=True, master=None):
        """Get the status of all the dashboards.

        Parameters
        ----------
        show_progress : Boolean, optional
            Show a dialog with progress, default is True
        """
        dashboards = self.dashboard_handler.dashboards
        if show_progress:
            dialog = tk.Toplevel(master=master)
            dialog.focus_set()  # set focus on the ProgressWindow
            dialog.grab_set()  # make a modal window
            dialog.transient(master)  # show only one window in the task bar

            dialog.title("Getting Status of Dashboards")
            dialog.resizable(False, False)  # window is not resizable
            # dialog.close gets fired when the window is destroyed
            # dialog.protocol(u'WM_DELETE_WINDOW', dialog.close)
            dialog.geometry("400x200+200+200")
            # cancel progress when <Escape> key is pressed
            # dialog.bind(u'<Escape>', self.close)

            progress = ttk.Progressbar(
                dialog,
                orient=tk.HORIZONTAL,
                length=len(dashboards) + 1,
                maximum=len(dashboards),
                mode="determinate",
                value=1,
            )
            progress.grid(ipady=20, sticky=tk.NSEW)
            label = ttk.Label(dialog, text="Dashboard")
            label.grid()
            dialog.rowconfigure(0, minsize=30)
            dialog.columnconfigure(0, weight=1)
            dialog.update_idletasks()
            dialog.after(50)

        result = []
        for dashboard in dashboards:
            if show_progress:
                label.configure({"text": dashboard})
                dialog.update_idletasks()
                dialog.after(50)
            status = self.status(dashboard)
            result.append((dashboard, status))
            if show_progress:
                progress.step()
                label.configure({"text": dashboard})
                dialog.update_idletasks()
                dialog.after(50)

        if show_progress:
            master.focus_set()
            dialog.destroy()

        return result

    def handle_add_dialog(self, result):
        """Handle the dialog to add a dashboard to the list."""
        save_dashboard = self.current_dashboard.name
        w = self["add"]
        dialog = w["dialog"]
        if result is None or result == "Cancel":
            dialog.deactivate(result)
        else:
            name = w["name"].get()
            url = w["url"].get()
            protocol = w["protocol"].get()

            if name in self.config:
                messagebox.showwarning(
                    "Duplicate name",
                    (
                        "There is already a dashboard called '{}'\n"
                        "Use a different name."
                    ).format(name),
                )
                return

            dialog.deactivate(result)

            # Now add to the configuration
            self.dashboard_handler.add_dashboard(name, url, protocol)

            # And reset the list in the dashboard combobox
            c = self["dashboard"]
            c.combobox.config({"value": self.dashboard_handler.dashboards})
            c.set(name)

            # check have credentials
            user, passwd = self.dashboard_handler.get_credentials(
                name, ask=self.ask_for_credentials
            )
            if user is None or passwd is None:
                # Unable to log in, so go back
                self["dashboard"].set(save_dashboard)

        dialog.destroy()
        del self["add"]

    def handle_dialog(self, result):
        """Handle the submit dialog being completed."""
        if result is None or result == "Cancel":
            self.dialog.deactivate(None)
        else:
            w = self
            self.dialog.deactivate(
                {
                    "project": w["project"].get(),
                    "title": w["title"].get(),
                    "dashboard": w["dashboard"].get(),
                    "description": w["description"].get(1.0, tk.END).strip("\n"),
                }
            )

    def handle_dashboard_dialog(self, result):
        """Handle the dialog to add a dashboard to the list."""
        w = self["display"]
        dialog = w["dialog"]
        dashboard = w["selected"].get()
        if result is None or result == "Cancel":
            dialog.deactivate(None)
        elif result == "Edit":
            self.edit_cb(dashboard)
        elif result == "Remove":
            self.remove_cb(dashboard)
        else:
            dialog.deactivate(None)

    def project_cb(self, event=None):
        """Handle a change in the project since it might be asking for adding a project
        in which case prompt for the new project's name, create it, and sleect it in the
        widget.
        """
        w = self
        project = w["project"].get()
        if project == "-- Create new project --":
            result = simpledialog.askstring("Add Project", "Project name:")
            if result is None or result == "":
                return
            # Add the project
            self.current_dashboard.add_project(result)
            self.dashboard_cb()
            w["project"].set(result)

    def reset_title(self):
        self["title"].set(self._flowchart.metadata["title"])

    def clear_title(self):
        self["title"].set("")

    def reset_description(self):
        self["description"].delete("1.0", "end")
        self["description"].insert("1.0", self._flowchart.metadata["description"])

    def clear_description(self):
        self["description"].delete("1.0", "end")

    def submit_with_dialog(self, flowchart):
        """
        Allow the user to choose the dashboard and other parameters,
        and submit the job as requested.

        Parameters
        ----------
        flowchart : seamm.Flowchart
            The flowchart to use.

        Returns
        -------
        job_id : integer
            The id of the submitted job.
        """
        self._flowchart = flowchart
        if self.dialog is None:
            title = flowchart.metadata["title"]
            description = flowchart.metadata["description"]
            self.create_submit_dialog(title=title, description=description)

        value = self._variable_value

        # Find any Parameter steps.
        parameter_steps = []
        step = flowchart.get_node("1")
        while step:
            if step.step_type == "control-parameters-step":
                parameter_steps.append(step)
            step = step.next()

        if len(parameter_steps) == 0:
            # Remove the parameter section
            self["parameters label"].grid_forget()
            self["parameters"].grid_forget()
            d = self.dialog.interior()
            d.rowconfigure(6, weight=0)
        else:
            self["parameters label"].grid(row=6, column=0, columnspan=2, sticky=tk.W)
            table = self["parameters"]
            table.clear()
            table.grid(row=7, column=0, columnspan=2, sticky=tk.NSEW)
            frame = table.interior()
            d = self.dialog.interior()
            d.rowconfigure(7, weight=1)
            row = 0
            for step in parameter_steps:
                variables = step.parameters["variables"]
                for name, data in variables.value.items():
                    table[row, 0] = name
                    if name not in value or value[name] is None:
                        value[name] = data["default"]
                    entry = ttk.Entry(frame)
                    entry.insert(0, value[name])
                    table[row, 1] = entry
                    table[row, 1].grid(sticky=tk.EW)
                    if data["type"] == "file":
                        button = tk.Button(
                            frame,
                            text="...",
                            command=(
                                lambda t=table, r=row, n=name, d=data: self.file_cb(
                                    t, r, n, d
                                )
                            ),
                        )
                        table[row, 2] = button
                    table[row, 3] = data["help"]
                    row += 1
            frame.columnconfigure(1, weight=1)

        # Post the dialog
        result = self.dialog.activate(geometry="centerscreenfirst")

        if result is not None:
            if len(parameter_steps) == 0:
                value = {}
            else:
                # Get the variable values
                table = self["parameters"]
                for row in range(table.nrows):
                    name = table[row, 0].cget("text")
                    value[name] = table[row, 1].get()

            dashboard_name = result.pop("dashboard")
            dashboard = self.dashboard_handler.get_dashboard(dashboard_name)
            job_id = dashboard.submit(flowchart, values=value, **result)
            return job_id
        else:
            return None



# -*- coding: utf-8 -*-

import logging
import pprint
import stevedore

"""A plugin manager based on Stevedore"""

logger = logging.getLogger(__name__)


class PluginManager(object):
    def __init__(self, namespace):
        logger.info("Initializing extensions for {}".format(namespace))

        self.namespace = namespace
        self.manager = stevedore.extension.ExtensionManager(
            namespace=self.namespace,
            invoke_on_load=True,
            on_load_failure_callback=self.load_failure,
        )
        self._plugins = {}

        logger.info(
            "Found {:d} extensions in '{:s}': {}".format(
                len(self.manager.names()), self.namespace, self.manager.names()
            )
        )

        logger.debug("Processing extensions")

        for name in self.manager.names():
            logger.debug("    extension name: {}".format(name))
            extension = self.manager[name].obj
            logger.debug("  extension object: {}".format(extension))
            data = extension.description()
            logger.debug("    extension data:")
            logger.debug(pprint.pformat(data))
            logger.debug("")
            group = data["group"]
            if group in self._plugins:
                self._plugins[group].append(name)
            else:
                self._plugins[group] = [name]

    def load_failure(self, mgr, ep, err):
        """Called when the extension manager can't load an extension"""
        logger.warning("Could not load %r: %s", ep.name, err)

    def get(self, name):
        return self.manager[name].obj

    def groups(self):
        return sorted(list(self._plugins.keys()))

    def plugins(self, group):
        return sorted(list(self._plugins[group]))



# -*- coding: utf-8 -*-

"""A node to split the flow in a flowchart"""

import logging
import seamm

logger = logging.getLogger(__name__)


class TkSplit(seamm.TkNode):
    """The Tk-based graphical representation of a splitting node"""

    anchor_points = {
        "n": (0, 0),
        "s": (0, 1),
        "e": (0.5, 0.5),
        "w": (-0.5, 0.5),
    }

    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=120, y=20, w=10, h=10
    ):
        """Initialize a node

        Keyword arguments:
        """
        logger.debug("Creating TkSplit, {} {} {} {} {}".format(node, x, y, w, h))
        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def draw(self):
        """Draw the node on the given canvas, making it visible"""

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y
        y1 = y0 + self.h
        self.border = self.canvas.create_oval(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, "type=outline"],
            fill="black",
        )



# -*- coding: utf-8 -*-

"""The ExecLocal object does what it name implies: it executes, or
runs, an executable locally."""

import humanize
import logging
import os
import os.path
from pathlib import Path
import platform
import pprint
import pyuca
import shutil
import stat
import subprocess
import sys
import tempfile
import time

if platform.system() != "Windows":
    import grp
    import pwd

logger = logging.getLogger(__name__)


class ExecLocal(object):
    def __init__(self):
        """Execute a flowchart, providing support for the actual
        execution of codes"""

        # times for formating 'ls' like output
        self.now = int(time.time())
        self.recent = self.now - (6 * 30 * 24 * 60 * 60)  # 6 months ago

    def run(
        self,
        cmd=[],
        input_data=None,
        files=None,
        env={},
        return_files=[],
        shell=False,
        in_situ=False,
        directory=None,
    ):
        """Execute 'cmd' in a temporary directory. 'files' is a dict
        keyed by filename of files to write before execution."""

        # Create temporary directory and write the files, being
        # careful about both errors and security.

        if directory is not None:
            directory = Path(directory)

        if in_situ:
            if directory is None:
                tmpdir = Path.cwd()
            else:
                tmpdir = Path(directory)
        else:
            tmpdir = Path(tempfile.mkdtemp())
            # Ensure the file is read/write by the creator only
            saved_umask = os.umask(0o077)

        logging.info(f"Running locally in {tmpdir}")

        if files is not None:
            # Write locally if directory is given
            if not in_situ and directory is not None:
                for filename in files:
                    path = directory / filename
                    mode = "wb" if type(files[filename]) is bytes else "w"
                    with path.open(mode=mode) as fd:
                        fd.write(files[filename])
            for filename in files:
                path = tmpdir / filename
                mode = "wb" if type(files[filename]) is bytes else "w"
                try:
                    with path.open(mode=mode) as fd:
                        fd.write(files[filename])
                except IOError:
                    logging.exception(
                        "An I/O error occured writing file '{}'".format(path)
                    )
                    if not in_situ:
                        os.umask(saved_umask)
                        shutil.rmtree(tmpdir)
                    return None
                except Exception:
                    logging.exception(
                        "An unexpected error occured writing file '{}'".format(path)
                    )
                    path.unlink()
                    if not in_situ:
                        os.umask(saved_umask)
                        shutil.rmtree(tmpdir)
                    return None
        # get a list of all existing files so we can determine what to delete
        if in_situ:
            existing = []
            existing_directories = []
            for dirpath, dirs, files in os.walk(tmpdir):
                dirpath = Path(dirpath)
                for name in files:
                    existing.append(dirpath / name)
                for name in dirs:
                    existing_directories.append(dirpath / name)

        # Now execute the program in the temp directory
        logger.debug("about to run " + " ".join(cmd))

        p = subprocess.run(
            cmd,
            input=input_data,
            env=dict(os.environ, **env),
            cwd=tmpdir,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
        )

        logger.debug("\n" + pprint.pformat(p))

        if not in_situ:
            os.umask(saved_umask)

        # capture the return code and output
        result = {
            "returncode": p.returncode,
            "stdout": p.stdout,
            "stderr": p.stderr,
        }

        # capture the list of files in the directory
        c = pyuca.Collator()
        listing = ""
        self.now = int(time.time())
        self.recent = self.now - (6 * 30 * 24 * 60 * 60)  # 6 months ago
        for dirpath, dirs, files in os.walk(tmpdir):
            # Do locale sensitive sort of files to list
            listing += (
                dirpath
                + "\n"
                + "\n\t".join(self.ls_format(dirpath, sorted(files, key=c.sort_key)))
            )
        result["listing"] = listing

        # capture the requested files
        result["files"] = []
        returned = []
        for pattern in return_files:
            paths = sorted(tmpdir.glob(pattern))
            for path in paths:
                returned.append(path)
                filename = path.name
                data = None
                exception = None
                result["files"].append(filename)
                try:
                    with path.open(mode="r") as fd:
                        data = fd.read()

                except UnicodeDecodeError:
                    with path.open(mode="rb") as fd:
                        data = fd.read()

                except IOError:
                    exception = sys.exc_info()
                    logging.warning(
                        f"An I/O error occurred reading file '{filename}'",
                        exc_info=exception,
                    )
                except Exception:
                    exception = sys.exc_info()
                    logging.warning(
                        f"An unexpected error occured reading file '{filename}'",
                        exc_info=exception,
                    )
                finally:
                    result[filename] = {"exception": exception, "data": data}

        # Clean up the temporary directory
        if in_situ:
            # Remove any files not originally here, or requested to return.
            for dirpath, dirs, files in os.walk(tmpdir):
                dirpath = Path(dirpath)
                for name in files:
                    filename = dirpath / name
                    if filename not in existing and filename not in returned:
                        filename.unlink()
                for name in dirs:
                    dirname = dirpath / name
                    if dirname not in existing_directories:
                        dirname.rmdir()
            # And move any files the need to go in subdirectories
            for filename in result["files"]:
                if filename[0] == "@":
                    subdir, fname = filename[1:].split("+")
                    from_path = directory / filename
                    subdir = directory / subdir
                    subdir.mkdir(parents=True, exist_ok=True)
                    from_path.rename(subdir / fname)
        else:
            shutil.rmtree(tmpdir)

            # And write the results locally
            if directory is not None:
                for filename in result["files"]:
                    if filename[0] == "@":
                        subdir, fname = filename[1:].split("+")
                        subdir = directory / subdir
                        subdir.mkdir(parents=True, exist_ok=True)
                        path = subdir / fname
                    else:
                        path = directory / filename

                    if result[filename]["data"] is not None:
                        mode = "wb" if type(result[filename]["data"]) is bytes else "w"
                        with path.open(mode=mode) as fd:
                            fd.write(result[filename]["data"])
                    else:
                        with path.open(mode="w") as fd:
                            fd.write(result[filename]["exception"])

        # Write any stdout and stderr

        if directory is not None:
            if "stdout" in result and result["stdout"] != "":
                path = directory / "stdout.txt"
                path.write_text(result["stdout"])
            if result["stderr"] != "":
                path = directory / "stderr.txt"
                path.write_text(result["stderr"])

        return result

    def get_mode_info(self, filename, mode):
        """Get the mode information for 'ls' like listing"""

        perms = "-"
        link = ""

        if stat.S_ISDIR(mode):
            perms = "d"
        elif stat.S_ISLNK(mode):
            perms = "l"
            link = os.readlink(filename)
        mode = stat.S_IMODE(mode)
        for who in "USR", "GRP", "OTH":
            for what in "R", "W", "X":
                # lookup attributes at runtime using getattr
                if mode & getattr(stat, "S_I" + what + who):
                    perms = perms + what.lower()
                else:
                    perms = perms + "-"
        # return multiple bits of info in a tuple
        return (perms, link)

    def ls_format(self, path, files):
        """Format a list of files as in 'ls'"""

        result = []
        for filename in files:
            try:  # exceptions
                # Get all the file info
                stat_info = os.lstat(os.path.join(path, filename))
            except Exception:
                result.append("{}: No such file or directory".format(filename))
                continue

            perms, link = self.get_mode_info(
                os.path.join(path, filename), stat_info.st_mode
            )

            nlink = "%4d" % stat_info.st_nlink  # formatting strings

            if platform.system() == "Windows":
                name = 8 * " "
                group = 8 * " "
            else:
                try:
                    name = "%-8s" % pwd.getpwuid(stat_info.st_uid)[0]
                except KeyError:
                    name = "%-8s" % stat_info.st_uid

                try:
                    group = "%-8s" % grp.getgrgid(stat_info.st_gid)[0]
                except KeyError:
                    group = "%-8s" % stat_info.st_gid

            size = humanize.naturalsize(stat_info.st_size)

            # Get time stamp of file
            ts = stat_info.st_mtime
            if (ts < self.recent) or (ts > self.now):  # boolean operators
                time_fmt = "%b %e  %Y"
            else:
                time_fmt = "%b %e %R"
            time_str = time.strftime(time_fmt, time.gmtime(ts))

            # Format the result
            tmp = "{} {} {} {} {} {} {}".format(
                perms, nlink, name, group, size, time_str, filename
            )

            if link:
                tmp += " -> " + link

            result.append(tmp)

        return result



# -*- coding: utf-8 -*-
"""Standard sets of parameters widely used in SEAMM.

Parameters
----------
structure_handling_parameters : dict(str, dict(str, str))
    Parameters for providing options for how to handle newly created structures.
    The options are:

        * Overwrite the current configuration in the current system.
        * Add a new configuration to the current system.
        * Create new system and a configuration in it to hold the structure.

    In addition, options are provided for naming the system and configuration whether
    or not new ones are created, i.e. the system and configuration can be renamed if
    they are reused.
"""


structure_handling_parameters = {
    "structure handling": {
        "default": "Overwrite the current configuration",
        "kind": "enum",
        "default_units": "",
        "enumeration": (
            "Overwrite the current configuration",
            "Create a new configuration",
            "Create a new system and configuration",
        ),
        "format_string": "s",
        "description": "First structure:",
        "help_text": (
            "Whether to overwrite the current configuration, or create a new "
            "configuration or system and configuration for the new structure"
        ),
    },
    "subsequent structure handling": {
        "default": "Create a new system and configuration",
        "kind": "enum",
        "default_units": "",
        "enumeration": (
            "Create a new configuration",
            "Create a new system and configuration",
        ),
        "format_string": "s",
        "description": "Subsequent structures:",
        "help_text": (
            "Whether to create a new configuration or system and configuration for "
            "new structures after the first structure, if any."
        ),
    },
    "system name": {
        "default": "keep current name",
        "kind": "string",
        "default_units": "",
        "enumeration": (
            "keep current name",
            "use SMILES string",
            "use Canonical SMILES string",
            "use IUPAC name",
            "use InChI",
            "use InChIKey",
        ),
        "format_string": "s",
        "description": "System name:",
        "help_text": "The name for the new system",
    },
    "configuration name": {
        "default": "use Canonical SMILES string",
        "kind": "string",
        "default_units": "",
        "enumeration": (
            "keep current name",
            "use SMILES string",
            "use Canonical SMILES string",
            "use IUPAC name",
            "use InChI",
            "use InChIKey",
        ),
        "format_string": "s",
        "description": "Configuration name:",
        "help_text": "The name for the new configuration",
    },
}


def structure_handling_description(P, **kwargs):
    """Return a standard description for how the structure will be handled.

    Parameters
    ----------
    P : dict(str, any)
        The dictionary of parameter values, which must contain the standard structure
        handling parameters.

    Returns
    -------
    str
        The text for printing.
    """

    text = ""

    handling = P["structure handling"]
    if handling == "Overwrite the current configuration":
        text += "The structure will overwrite the current configuration."
    elif handling == "Create a new configuration":
        text += "The structure will be put in a new configuration."
    elif handling == "Create a new system and configuration":
        text += "The structure will be put in a new system."
    else:
        raise ValueError(f"Do not understand how to handle the structure: '{handling}'")

    sysname = P["system name"]
    if sysname == "keep current name":
        text += " The name of the system will not be changed."
    elif sysname == "use SMILES string":
        text += " The name of the system will be its SMILES."
    elif sysname == "use Canonical SMILES string":
        text += " The name of the system will be its canonical SMILES."
    elif sysname == "use IUPAC name":
        text += " The name of the system will be its IUPAC name."
    elif sysname == "use InChI":
        text += " The name of the system will be its InChI."
    elif sysname == "use InChIKey":
        text += " The name of the system will be its InChIKey."
    else:
        tmp = safe_format(sysname, **kwargs)
        text += f" The name of the system will be '{tmp}'."

    confname = P["configuration name"]
    if confname == "keep current name":
        text += " The name of the configuration will not be changed."
    elif confname == "use SMILES string":
        text += " The name of the configuration will be its SMILES."
    elif confname == "use Canonical SMILES string":
        text += " The name of the configuration will be its canonical SMILES."
    elif confname == "use IUPAC name":
        text += " The name of the configuration will be its IUPAC name."
    elif confname == "use InChI":
        text += " The name of the configuration will be its InChI."
    elif confname == "use InChIKey":
        text += " The name of the configuration will be its InChIKey."
    else:
        tmp = safe_format(confname, **kwargs)
        text += f" The name of the configuration will be '{tmp}'."

    return text


def multiple_structure_handling_description(P, **kwargs):
    """Return a standard description for how the new structures will be handled.

    Parameters
    ----------
    P : dict(str, any)
        The dictionary of parameter values, which must contain the standard structure
        handling parameters.

    Returns
    -------
    str
        The text for printing.
    """

    text = "The first structure will "

    handling = P["structure handling"]
    if handling == "Overwrite the current configuration":
        text += "overwrite the current configuration."
    elif handling == "Create a new configuration":
        text += "be added as a new configuration of the current system."
    elif handling == "Create a new system and configuration":
        text += "be added as a new system and configuration."
    else:
        raise ValueError(f"Do not understand how to handle the structure: '{handling}'")

    handling = P["subsequent structure handling"]
    text += " Any subsequent structures will be "
    if handling == "Create a new configuration":
        text += "created as a new configuration of the current system."
    elif handling == "Create a new system and configuration":
        text += "created in a new system and configuration."
    else:
        raise ValueError(f"Do not understand how to handle the structure: '{handling}'")

    sysname = P["system name"]
    if sysname == "keep current name":
        text += " The name of the system will not be changed."
    elif sysname == "use SMILES string":
        text += " The name of the system will be its SMILES."
    elif sysname == "use Canonical SMILES string":
        text += " The name of the system will be its canonical SMILES."
    elif sysname == "use IUPAC name":
        text += " The name of the system will be its IUPAC name."
    elif sysname == "use InChI":
        text += " The name of the system will be its InChI."
    elif sysname == "use InChIKey":
        text += " The name of the system will be its InChIKey."
    else:
        tmp = safe_format(sysname, **kwargs)
        text += f" The name of the system will be '{tmp}'."

    confname = P["configuration name"]
    if confname == "keep current name":
        text += " The name of the configuration will not be changed."
    elif confname == "use SMILES string":
        text += " The name of the configuration will be its SMILES."
    elif confname == "use Canonical SMILES string":
        text += " The name of the configuration will be its canonical SMILES."
    elif confname == "use IUPAC name":
        text += " The name of the configuration will be its IUPAC name."
    elif confname == "use InChI":
        text += " The name of the configuration will be its InChI."
    elif confname == "use InChIKey":
        text += " The name of the configuration will be its InChIKey."
    else:
        tmp = safe_format(confname, **kwargs)
        text += f" The name of the configuration will be '{tmp}'."

    return text


def set_names(system, configuration, P, _first=True, **kwargs):
    """Set the names of the system and configuration.

    Parameters
    ----------
    system : _System
        The system being named

    configuration : _Configuration
        The configuration being named

    P : dict(str, any)
        The dictionary of parameter values, which must contain the standard structure
        handling parameters.

    _first : bool
        Whether this is the first or a subseqnet structure.

    kwargs : {str: str}
        keyword arguments providing values that may be substituted in the names.

    Returns
    -------
    str
        The text for printing.
    """
    sysname = P["system name"]
    if sysname == "keep current name":
        pass
    elif sysname == "use SMILES string":
        system.name = configuration.smiles
    elif sysname == "use Canonical SMILES string":
        system.name = configuration.canonical_smiles
    elif sysname == "use IUPAC name":
        system.name = configuration.PC_iupac_name(fallback=configuration.formula[0])
    elif sysname == "use InChI":
        system.name = configuration.inchi
    elif sysname == "use InChIKey":
        system.name = configuration.inchikey
    else:
        system.name = safe_format(sysname, **kwargs)

    confname = P["configuration name"]
    if confname == "keep current name":
        pass
    elif confname == "use SMILES string":
        configuration.name = configuration.smiles
    elif confname == "use Canonical SMILES string":
        configuration.name = configuration.canonical_smiles
    elif confname == "use IUPAC name":
        configuration.name = configuration.PC_iupac_name(
            fallback=configuration.formula[0]
        )
    elif confname == "use InChI":
        configuration.name = configuration.inchi
    elif confname == "use InChIKey":
        configuration.name = configuration.inchikey
    else:
        configuration.name = safe_format(confname, **kwargs)

    if _first:
        text = "The first structure "

        handling = P["structure handling"]
        if handling == "Overwrite the current configuration":
            text += "overwrote the current configuration, and was"
        elif handling == "Create a new configuration":
            text += "was added as a new configuration of the current system"
        elif handling == "Create a new system and configuration":
            text += "was added as a new system and configuration "
        else:
            raise ValueError(
                f"Do not understand how to handle the structure: '{handling}'"
            )
    else:
        handling = P["subsequent structure handling"]
        text = "This subsequent structure was "
        if handling == "Create a new configuration":
            text += "created as a new configuration of the current system"
        elif handling == "Create a new system and configuration":
            text += "created in a new system and configuration"
        else:
            raise ValueError(
                f"Do not understand how to handle the structure: '{handling}'"
            )

    text += f" named '{system.name}' / '{configuration.name}'."
    return text


def safe_format(s, *args, **kwargs):
    while True:
        try:
            return s.format(*args, **kwargs)
        except KeyError as e:
            e = e.args[0]
            kwargs[e] = "{%s}" % e



# -*- coding: utf-8 -*-

"""The start node in a flowchart"""

import copy
import pprint  # noqa: F401
import tkinter as tk
import tkinter.ttk as ttk

import seamm
import seamm_widgets as sw


class TkStartNode(seamm.TkNode):
    """The Tk-based graphical representation of a Start node"""

    anchor_points = {
        "s": (0, 0.5),
        "e": (0.5, 0.0),
        "w": (-0.5, 0.0),
    }

    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=150, y=50, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """
        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

        # Temporary storage for authors
        self._author_data = []

    def right_click(self, event):
        """Display the properties of the flowchart."""
        if self.popup_menu is not None:
            self.popup_menu.destroy()

        self.popup_menu = tk.Menu(self.canvas, tearoff=0)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

    def edit(self):
        """Present a dialog for editing this step's parameters.

        Subclasses can override this.
        """
        # Create the dialog if it doesn't exist
        if self.dialog is None:
            self.create_dialog()
            # After full creation, reset the dialog. This may do nothing,
            # or may layout the widgets, but can only be done after fully
            # creating the dialog.
            self.reset_dialog()
            # And resize the dialog to fit...
            self.fit_dialog()

        self.update_widgets()

        # And put it on-screen, the first time centered.
        result = self.dialog.activate(geometry="centerscreenfirst")

        if result == "OK":
            self.capture_metadata()
        self.cleanup_authors()

    def update_widgets(self):
        """Put the correct metadata into the widgets."""
        # Update the widgets with the current metadata from the flowchart
        metadata = self.tk_flowchart.flowchart.metadata

        self["title"].set(metadata["title"])
        self["description"].delete("1.0", tk.END)
        self["description"].insert("1.0", metadata["description"])
        self["keywords"].set(", ".join(metadata["keywords"]))
        if "grants" in metadata:
            self["grants"].set(", ".join(metadata["grants"]))

        # layout the authors
        self._author_data = copy.deepcopy(metadata["creators"])
        self.layout_authors()

    def cleanup_authors(self):
        """Destroy the internal widgets and reset the internal author data."""
        for data in self._author_data:
            for w in data["widgets"].values():
                w.destroy()
        self._author_data = []

    def capture_metadata(self):
        """Capture the metadata from the widgets and put to the flowchart."""
        metadata = self.tk_flowchart.flowchart.metadata

        metadata["title"] = self["title"].get()
        metadata["description"] = self["description"].get("1.0", tk.END)

        keywords = [x.strip() for x in self["keywords"].get().split(",")]
        if "seamm-flowchart" in keywords:
            keywords.remove("seamm-flowchart")
        metadata["keywords"] = keywords

        authors = []
        for data in self._author_data:
            w = data["widgets"]
            tmp = {}
            for item in ("name", "orcid", "affiliation"):
                value = w[item].get().strip()
                if value != "":
                    tmp[item] = value
            if len(tmp) > 0:
                authors.append(tmp)
        metadata["creators"] = authors

    def create_dialog(self):
        """Create a dialog for editing the flowchart properties."""
        frame = super().create_dialog("Flowchart Properties")

        f = self.create_frame(frame)
        f.grid(row=0, column=0, sticky=tk.NSEW)
        f.rowconfigure(0, weight=1)
        f.columnconfigure(0, weight=1)

    def create_frame(self, parent_widget):
        """Create a frame for editing the flowchart properties."""

        frame = self["inner frame"] = ttk.Frame(parent_widget)

        # Create the widgets and grid them in
        row = 0

        # title
        w = self["title"] = sw.LabeledEntry(frame, labeltext="Title:", width=100)
        w.grid(row=row, column=0, sticky=tk.EW)
        row += 1

        # description
        w = self["description label"] = ttk.Label(frame, text="Description")
        w.grid(row=row, column=0)
        row += 1

        w = self["description frame"] = sw.ScrolledFrame(
            frame, scroll_vertically=True, borderwidth=2, relief=tk.SUNKEN
        )
        w.grid(row=row, column=0, sticky=tk.NSEW)
        row += 1

        f = w.interior()
        w = self["description"] = tk.Text(f, wrap=tk.WORD, font=("Helvetica",))
        w.grid(sticky=tk.NSEW)
        f.rowconfigure(0, weight=1)
        f.columnconfigure(0, weight=1)

        # keywords
        w = self["keywords"] = sw.LabeledEntry(
            frame, labeltext="Keywords (comma separated):", width=100
        )
        w.grid(row=row, column=0, sticky=tk.EW)
        row += 1

        # grants
        w = self["grants"] = sw.LabeledEntry(
            frame, labeltext="Grants (DOIs, comma separated):", width=100
        )
        w.grid(row=row, column=0, sticky=tk.EW)
        row += 1

        # authors
        w = self["authors label"] = ttk.Label(frame, text="Authors")
        w.grid(row=row, column=0)
        row += 1

        w = self["authors frame"] = sw.ScrolledFrame(
            frame, scroll_vertically=True, borderwidth=2, relief=tk.SUNKEN
        )
        w.grid(row=row, column=0, sticky=tk.NSEW)
        row += 1

        f = w.interior()
        self["name label"] = ttk.Label(f, text="Name (last, others)")
        self["orcid label"] = ttk.Label(f, text="ORCID")
        self["affiliation label"] = ttk.Label(f, text="Affiliation")

        frame.columnconfigure(0, weight=1)

        return frame

    def draw(self):
        """Draw the node on the given canvas, making it visible"""

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y - self.h / 2
        y1 = y0 + self.h
        self.border = self.canvas.create_oval(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, "type=outline"],
            fill="white",
        )

        # the label in the middle
        self.title_label = self.canvas.create_text(
            self.x, self.y, text=self.title, tags=[self.tag, "type=title"]
        )

    def handle_dialog(self, result):
        self.dialog.deactivate(result)

    def layout_authors(self):
        """Layout the table of authors."""

        frame = self["authors frame"].interior()

        # Unpack any widgets
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Put in the column headers.
        row = 0
        self["name label"].grid(row=row, column=1)
        self["orcid label"].grid(row=row, column=2)
        self["affiliation label"].grid(row=row, column=3)
        row += 1

        for data in self._author_data:
            name = data["name"]
            orcid = data.get("orcid", "")
            affiliation = data.get("affiliation", "")
            if "widgets" not in data:
                widgets = data["widgets"] = {}
            else:
                widgets = data["widgets"]

            if "remove" not in widgets:
                # The button to remove a row...
                widgets["remove"] = ttk.Button(
                    frame,
                    text="-",
                    width=2,
                    command=lambda row=row: self.remove_author(row),
                    takefocus=True,
                )

            if "name" not in widgets:
                # the authors name
                widgets["name"] = ttk.Entry(frame, width=50, takefocus=True)
                widgets["name"].insert("end", name)

            if "orcid" not in widgets:
                # The author's ORCID id
                widgets["orcid"] = ttk.Entry(frame, width=20, takefocus=True)
                widgets["orcid"].insert("end", orcid)

            if "affiliation" not in widgets:
                # The author's affiliation
                widgets["affiliation"] = ttk.Entry(frame, width=50, takefocus=True)
                widgets["affiliation"].insert("end", affiliation)

            widgets["remove"].grid(row=row, column=0, sticky=tk.W)
            widgets["name"].grid(row=row, column=1, stick=tk.EW)
            widgets["orcid"].grid(row=row, column=2, stick=tk.EW)
            widgets["affiliation"].grid(row=row, column=3, stick=tk.EW)

            row += 1

        # The button to add a row...
        if "add author" not in self:
            w = self["add author"] = ttk.Button(
                frame,
                text="+",
                width=5,
                command=self.add_author,
                takefocus=True,
            )
            w.focus_set()
        else:
            w = self["add author"]
        w.lift()
        w.grid(row=row, column=0, columnspan=3, sticky=tk.W)

        frame.grid_columnconfigure(3, weight=1)

    def add_author(self):
        """Add a new row to the author table."""
        self._author_data.append({"name": "", "orcid": "", "affiliation": ""})
        self.layout_authors()

    def remove_author(self, row):
        """Remove a author entry from the table.

        Parameters
        ----------
        row : int
            The row in the table to remove. Note the first author is at row 1.
        """
        index = row - 1
        data = self._author_data[index]
        if "widgets" in data:
            for w in data["widgets"].values():
                w.destroy()
        del self._author_data[index]

        self.layout_authors()



# -*- coding: utf-8 -*-

"""The ExecWorkflow object does what it name implies: it executes, or
runs, a given flowchart.

It provides the environment for running the computational tasks
locally or remotely, using what is commonly called workflow management
system (WMS).  The WMS concept, as used here, means tools that run
given tasks without knowing anything about chemistry. The chemistry
specialization is contained in the Flowchart and the nodes that it
contains."""

import calendar
import logging
import os.path
from pathlib import Path
import string
import sys
import traceback

from molsystem import SystemDB
import reference_handler
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __  # noqa: F401
from seamm_util import getParser

logger = logging.getLogger(__name__)
job = printing.getPrinter()


class ExecFlowchart(object):
    def __init__(self, flowchart=None):
        """Execute a flowchart, providing support for the actual
        execution of codes"""
        logger.info("In ExecFlowchart.init()")

        self.flowchart = flowchart

    def run(self, root=None):
        logger.info("In ExecFlowchart.run()")
        if not self.flowchart:
            raise RuntimeError("There is no flowchart to run!")

        # Get the command line options
        parser = getParser(name="SEAMM")
        options = parser.get_options()

        # Set the options in each step
        for node in self.flowchart:
            node.all_options = options

        # Create the global context
        logger.info("Creating global variables space")
        seamm.flowchart_variables = seamm.Variables()

        # And add the printer
        seamm.flowchart_variables.set_variable("printer", job)

        # Setup the citations
        filename = Path(self.flowchart.root_directory) / "references.db"
        filename.unlink(missing_ok=True)
        references = None
        try:
            references = reference_handler.Reference_Handler(str(filename))
        except Exception as e:
            job.job("Error with references:")
            job.job(e)

        if references is not None:
            template = string.Template(
                """\
                @misc{seamm,
                  address      = {Virginia Tech, Blacksburg, VA, USA},
                  author       = {Jessica Nash and
                                  Eliseo Marin-Rimoldi and
                                  Mohammad Mostafanejad and
                                  Paul Saxe},
                  doi          = {10.5281/zenodo.5153984},
                  month        = {$month},
                  note         = {Funding: NSF OAC-1547580 and CHE-2136142},
                  organization = {The Molecular Sciences Software Institute (MolSSI)},
                  publisher    = {Zenodo},
                  title        = {SEAMM: Simulation Environment for Atomistic and
                                  Molecular Modeling},
                  url          = {https://doi.org/10.5281/zenodo.5153984},
                  version      = {$version},
                  year         = $year
                }"""
            )

            try:
                version = seamm.__version__
                year, month = version.split(".")[0:2]
                month = calendar.month_abbr[int(month)].lower()
                citation = template.substitute(
                    month=month,
                    version=version,
                    year=year,
                )

                references.cite(
                    raw=citation,
                    alias="SEAMM",
                    module="seamm",
                    level=1,
                    note="The principle citation for SEAMM.",
                )
            except Exception as e:
                job.job(f"Exception in citation {type(e)}: {e}")
                job.job(traceback.format_exc())

        # Create the system database, default system and configuration
        if "SEAMM" in options:
            seamm_options = options["SEAMM"]
            read_only = "read_only" in seamm_options and seamm_options["read_only"]
            db_file = seamm_options["database"]
            if ":memory:" in db_file:
                db = SystemDB(filename=db_file)
            else:
                path = Path(db_file).expanduser().resolve()
                uri = "file:" + str(path)
                if read_only:
                    uri += "?mode=ro"
                db = SystemDB(filename=uri)
        else:
            db = SystemDB(filename="file:seamm.db")

        # Put the system database in the global context for access.
        seamm.flowchart_variables.set_variable("_system_db", db)

        self.flowchart.root_directory = root

        # Correctly number the nodes
        self.flowchart.set_ids()

        # Write out an initial summary of the flowchart before doing anything
        # Reset the visited flag for traversal
        self.flowchart.reset_visited()

        # Get the start node
        next_node = self.flowchart.get_node("1")

        # describe ourselves
        job.job(("\nDescription of the flowchart" "\n----------------------------"))

        while next_node:
            # and print the description
            try:
                next_node = next_node.describe()
            except Exception:
                message = "Error describing the flowchart\n\n" + traceback.format_exc()
                print(message)
                logger.critical(message)
                raise
            except:  # noqa: E722
                message = (
                    "Unexpected error describing the flowchart\n\n"
                    + traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise

        job.job("")

        # And actually run it!
        job.job(("Running the flowchart\n" "---------------------"))

        try:
            next_node = self.flowchart.get_node("1")
            while next_node is not None:
                try:
                    next_node = next_node.run()
                except DeprecationWarning as e:
                    print("\nDeprecation warning: " + str(e))
                    traceback.print_exc(file=sys.stderr)
                    traceback.print_exc(file=sys.stdout)
        finally:
            # Write the final structure
            db = seamm.flowchart_variables.get_variable("_system_db")
            system = db.system
            if system is not None:
                configuration = system.configuration
                if configuration is not None:
                    output = []
                    if configuration.n_atoms > 0:
                        # MMCIF file has bonds
                        filename = os.path.join(
                            self.flowchart.root_directory, "final_structure.mmcif"
                        )
                        text = None
                        try:
                            text = configuration.to_mmcif_text()
                        except Exception:
                            pass
                        if text is not None:
                            with open(filename, "w") as fd:
                                print(text, file=fd)
                            output.append("final_structure.mmcif")

                        # CIF file has cell and may have bonds
                        if configuration.periodicity == 3:
                            text = None
                            try:
                                text = configuration.to_cif_text()
                            except Exception:
                                pass
                            if text is not None:
                                filename = os.path.join(
                                    self.flowchart.root_directory, "final_structure.cif"
                                )
                                with open(filename, "w") as fd:
                                    print(configuration.to_cif_text(), file=fd)
                                    output.append("final_structure.cif")
                        if len(output) > 0:
                            files = "' and '".join(output)
                            job.job(
                                f"\nWrote the final structure to '{files}' for viewing."
                            )
                        else:
                            job.job(
                                "\nWas unable to write the final structure as either "
                                "an mmcif or cif file for viewing."
                            )

            # And print out the references
            filename = os.path.join(self.flowchart.root_directory, "references.db")
            try:
                references = reference_handler.Reference_Handler(filename)
            except Exception as e:
                job.job("Error with references:")
                job.job(e)

            if references.total_citations() > 0:
                tmp = {}
                citations = references.dump(fmt="text")
                for citation, text, count, level in citations:
                    if level not in tmp:
                        tmp[level] = {}
                    tmp[level][citation] = (text, count)

                n = 0
                for level in sorted(tmp.keys()):
                    ref_dict = tmp[level]
                    if level == 1:
                        job.job("\nPrimary references:\n")
                        n = 0
                    elif level == 2:
                        job.job("\nSecondary references:\n")
                        n = 0
                    else:
                        job.job("\nLess important references:\n")
                        n = 0

                    lines = []
                    for citation in sorted(ref_dict.keys()):
                        n += 1
                        text, count = ref_dict[citation]
                        if count == 1:
                            lines.append("({}) {:s}".format(n, text))
                        else:
                            lines.append(
                                "({}) {:s} (used {:d} times)".format(n, text, count)
                            )
                    job.job(
                        __("\n\n".join(lines), indent=4 * " ", indent_initial=False)
                    )



# -*- coding: utf-8 -*-

"""The base class for Tk nodes (steps) in the GUI for flowcharts."""

import collections.abc
import copy
import logging
import json
import Pmw
import seamm
from seamm_util import default_units
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)


class TkNode(collections.abc.MutableMapping):
    """The base class for Tk nodes (steps) in the GUI for flowcharts.

    Parameters
    ----------
    tk_flowchart : seamm.TkFlowchart
        The graphical flowchart this step is in.
    node : seamm.Node
        The non-graphical node this corresponds to.
    node_type : "simple", "loop"
        The type of node on the graph. "simple" has an in and out arrow. "loop" has
        three arrows.
    canvas: tkinter.Canvas
        The Canvas widget that this node is drawn on.
    x: int
        The x-coordinate the drawing for the node on the canvas.
    y: int
        The y-coordinate of the drawing for the node on the canvas.
    w: int
        The width of the drawing for the node on the canvas.
    h: int
        The height of the drawing for the node on the canvas.
    my_logger : logging.Logger, optional
        The logger to use. Defaults to the global one defined in the module.

    Fields
    ------
    border
    canvas
    dialog : tkinter.Toplevel
        The dialog for editing the parameters.
    flowchart
    h
    logger : logging.Logger
        The logger for debug & warning output.
    node : seamm.Node
        The non-graphical node corresponding to this graphical one.
    node_type : enum("simple", "loop")
        The type of the node from the point of connectivity.
    popup_menu : tkinter.Menu
        The popup menu used for right-clicks.
    selected
    tag
    title
    title_label : tkinter.ttk.Label
        The label for the title of the step on the display.
    tk_flowchart : seamm.TkFlowchart
        The Tk Flowchart that contains this step.
    tk_subflowchart : seamm.TkFlowchart
        The sub flowchart is if this step contains one.
    w
    x
    y
    uuid

    Notes
    -----
    The state is held in the corresponding non-graphical node, `self.node`. Many of
    the properties are thin-wrappers to the same property of the non-graphical node.
    """

    anchor_points = {
        "s": (+0.00, +0.50),
        "sse": (+0.25, +0.50),
        "se": (+0.50, +0.50),
        "ese": (+0.50, +0.25),
        "e": (+0.50, +0.00),
        "ene": (+0.50, -0.25),
        "ne": (+0.50, -0.50),
        "nne": (+0.25, -0.50),
        "n": (+0.00, -0.50),
        "nnw": (-0.25, -0.50),
        "nw": (-0.50, -0.50),
        "wnw": (-0.50, -0.25),
        "w": (-0.50, +0.00),
        "wsw": (-0.50, +0.25),
        "sw": (-0.50, +0.50),
        "ssw": (-0.25, +0.50),
    }

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        node_type="simple",
        canvas=None,
        x=None,
        y=None,
        w=None,
        h=None,
        my_logger=logger,
    ):
        """Initialize a node

        Keyword arguments:
        """
        self._border = None
        self._selected = False
        self._tmp = None
        self._canvas = None

        self.canvas = canvas

        self.dialog = None
        self.logger = my_logger
        self.node = node
        self.node_type = node_type
        self.popup_menu = None
        self._tables = None  # Temporary list of all tables up to an including this node
        self.title_label = None
        self.tk_flowchart = tk_flowchart
        self.tk_subflowchart = None

        if self.node is not None:
            if self.node.x is None:
                self.node.x = x
            if self.node.y is None:
                self.node.y = y
            if self.node.w is None:
                self.node.w = w
            if self.node.h is None:
                self.node.h = h

        # Widget information
        self._widget = {}
        self.tk_var = {}
        self.results_widgets = None

        # Because the default for saving properties in the database is True
        # we need to initialize the results to include them by default
        if self.node.parameters is not None and "results" in self.node.parameters:
            self.initialize_results()

    def __hash__(self):
        """Provide a unique key to make iterable."""
        return self.node.uuid

    def __eq__(self, other):
        """Test for equality (identity) with another node."""
        return self.__class__ == other.__class__ and self.__hash__() == other.__hash__()

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets."""
        return self._widget[key]

    def __setitem__(self, key, value):
        """Allow [key] access to set a widget."""
        self._widget[key] = value

    def __delitem__(self, key):
        """Allow deletion of widgets."""
        if key in self._widget:
            self._widget[key].destroy()
        del self._widget[key]

    def __iter__(self):
        """Allow iteration over the widgets"""
        return iter(self._widget)

    def __len__(self):
        """Provide the nmber of widgets, for e.g. len() command."""
        return len(self._widget)

    @property
    def border(self):
        """The border of the picture in the flowchart"""
        return self._border

    @border.setter
    def border(self, value):
        self._border = value

    @property
    def canvas(self):
        """The canvas for drawing the node"""
        return self._canvas

    @canvas.setter
    def canvas(self, value):
        self._canvas = value

    @property
    def flowchart(self):
        """The flowchart object"""
        return self.node.flowchart

    @flowchart.setter
    def flowchart(self, value):
        """The flowchart object"""
        self.node.flowchart = value

    @property
    def h(self):
        """The height of the graphical node"""
        return self.node.h

    @h.setter
    def h(self, value):
        self.node.h = value

    @property
    def selected(self):
        """Whether I am selected or not"""
        return self._selected

    @selected.setter
    def selected(self, value):
        self._selected = value
        if value:
            self.canvas.itemconfigure(self.border, outline="red")
        else:
            self.canvas.itemconfigure(self.border, outline="black")

    @property
    def tag(self):
        """The string representation of the uuid of the node"""
        return self.node.tag

    @property
    def title(self):
        """The title to display"""
        return self.node.title

    @title.setter
    def title(self, value):
        self.node.title = value
        if self.title_label is not None:
            self.canvas.itemconfigure(self.title_label, text=value)

    @property
    def w(self):
        """The width of the graphical node"""
        return self.node.w

    @w.setter
    def w(self, value):
        self.node.w = value

    @property
    def x(self):
        """The x-position of the center of the graphical node"""
        return self.node.x

    @x.setter
    def x(self, value):
        self.node.x = value

    @property
    def y(self):
        """The y-position of the center of the graphical node"""
        return self.node.y

    @y.setter
    def y(self, value):
        self.node.y = value

    @property
    def uuid(self):
        """The uuid of the node"""
        return self.node.uuid

    def activate(self):
        """Add active handles at the anchor points and change the cursor."""

        self.canvas.delete(self.tag + " && type=anchor")
        for pt, x, y in self.anchor_point("all"):
            x0 = x - 2
            y0 = y - 2
            x1 = x + 2
            y1 = y + 2
            self.canvas.create_oval(
                x0,
                y0,
                x1,
                y1,
                fill="red",
                outline="red",
                tags=[self.tag, "type=anchor", "anchor=" + pt],
            )

    def activate_anchor_point(self, point, halo):
        """Put a marker on the anchor point to indicate it is under the cursor."""

        x, y = self.anchor_point(point)
        self.canvas.create_oval(
            x - halo,
            y - halo,
            x + halo,
            y + halo,
            fill="red",
            outline="red",
            tags=[self.tag, "type=active_anchor", "anchor=" + point],
        )

    def anchor_point(self, anchor="all"):
        """Where the anchor points are located. If "all" is given
        a dictionary of all points is returned"""

        if anchor == "all":
            result = []
            for pt in type(self).anchor_points:
                a, b = type(self).anchor_points[pt]
                result.append((pt, int(self.x + a * self.w), int(self.y + b * self.h)))
            return result

        if anchor in type(self).anchor_points:
            a, b = type(self).anchor_points[anchor]
            return (int(self.x + a * self.w), int(self.y + b * self.h))

        raise NotImplementedError("anchor position '{}' not implemented".format(anchor))

    def check_anchor_points(self, x, y, halo):
        """If the position x, y is within halo or one of the anchor points
        activate the point and return the name of the anchor point
        """

        points = []
        for direction, edge in self.connections():
            if direction == "out":
                points.append(edge.anchor1)
            else:
                points.append(edge.anchor2)

        for point, x0, y0 in self.anchor_point():
            if x >= x0 - halo and x <= x0 + halo and y >= y0 - halo and y <= y0 + halo:
                if point in points:
                    return None
                else:
                    return point
        return None

    def connections(self):
        """Return a list of all the incoming and outgoing edges
        for this node, giving the anchor points and other node
        """

        return self.tk_flowchart.edges(self)

    def create_dialog(
        self,
        title="Edit step",
        widget="frame",
        results_tab=False,
    ):
        """Create the base dialog for editing the parameters for a step.

        Parameters
        ----------
        title : str
            The title of the dialog.
        widget : enum
            Whether to use a simple dialog ("frame") or use a notebook ("notebook").
        results_tab : bool
            **OBSOLETE** Not longer used.
        """
        toplevel = self.canvas.winfo_toplevel()

        self.logger.debug("Create dialog in tk_node base class")
        self.dialog = Pmw.Dialog(
            toplevel,
            buttons=("OK", "Cancel"),
            master=toplevel,
            title=title,
            command=self.handle_dialog,
        )
        self.dialog.withdraw()

        results_tab = (
            self.node.parameters is not None and "results" in self.node.parameters
        )

        if widget == "notebook" or results_tab or "keywords" in self.node.metadata:
            # A tabbed notebook
            notebook = ttk.Notebook(self.dialog.interior())
            notebook.pack(side="top", fill=tk.BOTH, expand=tk.YES)
            self["notebook"] = notebook

            # Main frame holding the widgets
            frame = ttk.Frame(notebook)
            self["frame"] = frame
            notebook.add(frame, text="Parameters", sticky=tk.NSEW)
        elif widget == "frame":
            # Create a frame to hold everything
            frame = ttk.Frame(self.dialog.interior())
            frame.pack(expand=tk.YES, fill=tk.BOTH)
            self["frame"] = frame
            return frame

        if results_tab:
            # Second tab for results if requested
            rframe = self["results frame"] = ttk.Frame(notebook)
            notebook.add(rframe, text="Results", sticky=tk.NSEW)

            # Shortcut for parameters
            P = self.node.parameters

            if "create tables" in P:
                var = self.tk_var["create tables"] = tk.IntVar()
                if P["create tables"].value == "yes":
                    var.set(1)
                else:
                    var.set(0)
                self["create tables"] = ttk.Checkbutton(
                    rframe, text="Create tables if needed", variable=var
                )
                self["create tables"].grid(row=0, column=0, sticky=tk.W)

            self["results"] = sw.ScrolledColumns(
                rframe,
                columns=[
                    "Result",
                    "Save in database",
                    "",
                    "Save in variable named",
                    "Save in table",
                    "as Column name",
                    "Units",
                ],
            )
            self["results"].grid(row=1, column=0, sticky=tk.NSEW)
            rframe.columnconfigure(0, weight=1)
            rframe.rowconfigure(1, weight=1)

        if "keywords" in self.node.metadata:
            # Next tab to handle adding keywords manually
            self.logger.debug("Adding the keyword tab")
            kframe = self["add_to_input"] = ttk.Frame(notebook)
            notebook.add(kframe, text="Add to input", sticky=tk.NSEW)
            self["keywords"] = sw.Keywords(
                kframe,
                metadata=self.node.metadata["keywords"],
                keywords=self.node.parameters["extra keywords"].value,
            )
            self["keywords"].pack(expand="yes", fill="both")

        return frame

    def deactivate(self):
        """Remove the decorations that indicate active anchor points"""

        self.canvas.delete(self.tag + " && type=anchor")
        self.canvas.delete(self.tag + " && type=active_anchor")

    def default_edge_subtype(self):
        """Return the default subtype of the edge. Usually this is ''
        but for nodes with two or more edges leaving them, such as a loop, this
        method will return an appropriate default for the current edge. For
        example, by default the first edge emanating from a loop-node is the
        'loop' edge; the second, the 'exit' edge.

        A return value of 'too many' indicates that the node exceeds the number
        of allowed exit edges.
        """

        # how many outgoing edges are there?
        n_edges = len(self.tk_flowchart.edges(self, direction="out"))

        self.logger.debug("node.default_edge_label, n_edges = {}".format(n_edges))

        if n_edges == 0:
            return "next"
        else:
            return "too many"

    def double_click(self, event):
        """Handle a double-click on the node.

        This method raises the dialog to edit the parameters.  Subclasses should
        override this as appropriate!
        """

        self.edit()

    def draw(self):
        """Draw the node on the given canvas, making it visible"""
        # Remove any graphics items
        self.undraw()

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y - self.h / 2
        y1 = y0 + self.h
        self.border = self.canvas.create_rectangle(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, "type=outline"],
            fill="white",
        )

        # the label in the middle
        self.title_label = self.canvas.create_text(
            self.x, self.y, text=self.title, tags=[self.tag, "type=title"]
        )

        for direction, edge in self.connections():
            edge.move()

    def edit(self):
        """Present a dialog for editing this step's parameters.

        Subclasses can override this.
        """
        # Create the dialog if it doesn't exist
        if self.dialog is None:
            self.create_dialog()
            # After full creation, reset the dialog. This may do nothing,
            # or may layout the widgets, but can only be done after fully
            # creating the dialog.
            self.reset_dialog()
            # And resize the dialog to fit...
            self.fit_dialog()

        # And put it on-screen, the first time centered. If it contains
        # a subflowchart, save it so it can be restored on a 'Cancel'
        if self.tk_subflowchart is not None:
            self.tk_subflowchart.push()

        self.dialog.activate(geometry="centerscreenfirst")

    def end_move(self, deltax, deltay):
        """End moving the node on the canvas.

        Parameters
        ----------
        deltax : int
            The number of pixels to move in the x-direction.
        deltay : int
            The number of pixels to move in the y-direction.
        """
        self.move(deltax, deltay)
        self._x0 = None
        self._y0 = None
        self._tmp = None

    def fit_dialog(self):
        """Resize and fit the dialog to the current contents and the
        constraint of the window.
        """
        self.logger.debug("Entering fit_dialog")
        frame = self["frame"]
        frame.update_idletasks()
        width = frame.winfo_width()
        height = frame.winfo_height()
        sw = frame.winfo_screenwidth()
        sh = frame.winfo_screenheight()

        self.logger.debug(
            "  frame wxh = {} x {}, screen = {} x {}".format(width, height, sw, sh)
        )

        mw = 0
        mh = 0
        if "notebook" in self:
            for tab in self["notebook"].tabs():
                widget = frame.nametowidget(tab)
                widget.update_idletasks()
                self.logger.debug("  widget = {}".format(widget))
                ww = widget.winfo_width()
                hh = widget.winfo_height()
                w = widget.winfo_reqwidth()
                h = widget.winfo_reqheight()
                self.logger.debug(
                    "  tab {} wxh = {} x {}, requested = {} x {}".format(
                        tab, ww, hh, w, h
                    )
                )
                if w > mw:
                    mw = w
                if h > mh:
                    mh = h
                if ww > width:
                    width = ww
                if hh > height:
                    height = hh
            # Need to do results again using the inside of the scrolled table..
            if "results" in self:
                widget = self["results"].interior()
                self.logger.debug("  widget = {}".format(widget))
                widget.update_idletasks()
                ww = widget.winfo_width()
                hh = widget.winfo_height()
                w = widget.winfo_reqwidth()
                h = widget.winfo_reqheight()
                self.logger.debug(
                    "  tab {} wxh = {} x {}, requested = {} x {}".format(
                        tab, ww, hh, w, h
                    )
                )
                if w > mw:
                    mw = w
                if h > mh:
                    mh = h
                if ww > width:
                    width = ww
                if hh > height:
                    height = hh
        else:
            mw = frame.winfo_reqwidth()
            mh = frame.winfo_reqheight()
            self.logger.debug("  frame requested = {} x {}".format(mw, mh))

        if width < mw:
            width = mw
        width += 70
        if width + 70 > 0.9 * sw:
            width = int(0.9 * sw)
        if height < mh:
            height = mh
        height += 70
        if height > 0.9 * sh:
            height = int(0.9 * sh)

        self.dialog.geometry("{}x{}".format(width, height))

    def from_flowchart(self, tk_flowchart=None, flowchart=None):
        """Recreate the graphics from the non-graphical flowchart.
        Only used in nodes that contain flowchart"""

        if self.tk_subflowchart is None or self.node.subflowchart is None:
            return

        self.tk_subflowchart.clear()

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in self.node.subflowchart:
            extension = node.extension
            if extension is None:
                # Start node
                translate[node] = self.tk_subflowchart.get_node("1")
            else:
                new_node = copy.copy(node)
                self.logger.debug("creating {} node".format(extension))
                plugin = self.tk_subflowchart.plugin_manager.get(extension)
                self.logger.debug("  plugin object: {}".format(plugin))
                tk_node = plugin.create_tk_node(
                    tk_flowchart=self.tk_subflowchart,
                    canvas=self.tk_subflowchart.canvas,
                    node=new_node,
                )
                translate[node] = tk_node
                tk_node.from_flowchart()
                self.tk_subflowchart.graph.add_node(tk_node)
                tk_node.draw()

        # And the edges
        for edge in self.node.subflowchart.edges():
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            attr = {}
            for key in edge:
                if key not in ("node1", "node2"):
                    attr[key] = edge[key]
            self.tk_subflowchart.add_edge(node1, node2, **attr)

    def handle_dialog(self, result):
        """Do the right thing when the dialog is closed."""
        if result is None or result == "Cancel":
            self.dialog.deactivate(result)

            # If there is a subflowchart, revert to the saved copy
            if self.tk_subflowchart is not None:
                self.tk_subflowchart.pop()

            # Reset the results widgets if they exist
            if self.results_widgets is not None:
                results = self.node.parameters["results"]["value"]
                self.logger.debug("Resetting results on Cancel")
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug("  results dict\n---------")
                    for key, item in results.items():
                        self.logger.debug(key)
                        self.logger.debug(
                            json.dumps(results[key], sort_keys=True, indent=3)
                        )

                for (
                    key,
                    w_property,
                    w_check,
                    w_variable,
                    w_table,
                    w_column,
                    w_units,
                ) in self.results_widgets:  # noqa: E501
                    self.logger.debug("  key: {}".format(key))
                    w_variable.delete(0, tk.END)
                    if w_table is not None:
                        w_table.delete(0, tk.END)
                    if w_column is not None:
                        w_column.delete(0, tk.END)
                    if key in results:
                        tmp = results[key]
                        self.logger.debug(
                            "  key dict\n---------\n"
                            + json.dumps(tmp, sort_keys=True, indent=3)
                            + "\n-----"
                        )
                        if "variable" in tmp:
                            self.tk_var[key].set(1)
                            w_variable.insert(0, tmp["variable"])
                        else:
                            self.tk_var[key].set(0)
                            w_variable.insert(0, key.lower().replace(" ", "_"))

                        if w_table is not None:
                            if "table" in tmp:
                                w_table.insert(0, tmp["table"])
                                w_column.insert(0, tmp["column"])
                            else:
                                w_table.set("")
                                w_column.insert(0, key.lower().replace("_", " "))

                        if w_property is not None:
                            if "property" in tmp:
                                self.tk_var["property " + key]["value"].set(1)
                            else:
                                self.tk_var["property " + key]["value"].set(0)

                        if w_units is not None:
                            if "units" in tmp:
                                w_units.set(tmp["units"])
                    else:
                        self.logger.debug("  resetting widgets")
                        self.tk_var[key].set(0)
                        w_variable.insert(0, key.lower().replace(" ", "_"))
                        if w_column is not None:
                            w_column.insert(0, key.lower().replace("_", " "))

            # Reset the parameters, if any
            if self.node.parameters is not None:
                self.node.parameters.reset_widgets()

            # Reset any keywords
            if "keywords" in self:
                self["keywords"].reset()

            # Reset the layout to make sure it is correct
            self.reset_dialog()

        elif result == "Help":
            self.help()
        elif result == "OK":
            self.dialog.deactivate(result)

            # Capture the parameters from the widgets
            if self.node.parameters is not None:
                self.node.parameters.set_from_widgets()

            # If there is a subflowchart, throw the saved copy away
            if self.tk_subflowchart is not None:
                self.tk_subflowchart.pop_and_discard()

            # Get what results to store, if the results tab exists
            if self.results_widgets is not None:
                # Shortcut for parameters
                P = self.node.parameters

                # and from the results tab...
                if "create tables" in P:
                    if self.tk_var["create tables"].get():
                        P["create tables"].value = "yes"
                    else:
                        P["create tables"].value = "no"

                results = P["results"].value = {}
                for (
                    key,
                    w_property,
                    w_check,
                    w_variable,
                    w_table,
                    w_column,
                    w_units,
                ) in self.results_widgets:  # noqa: E501
                    tmp = {}
                    if self.tk_var[key].get():
                        tmp["variable"] = w_variable.get()
                        if w_units is not None:
                            tmp["units"] = w_units.get()
                    if w_table is not None:
                        table = w_table.get()
                        if table != "":
                            tmp["table"] = table
                            tmp["column"] = w_column.get()
                            if w_units is not None:
                                tmp["units"] = w_units.get()
                    if w_property is not None:
                        if self.tk_var["property " + key]["value"].get() == 1:
                            tmp["property"] = self.tk_var["property " + key]["key"]
                        elif "property" in tmp:
                            del tmp["property"]
                    if len(tmp) > 0:
                        results[key] = tmp

            # And any keywords
            if "keywords" in self:
                P = self.node.parameters
                P["extra keywords"].value = self["keywords"].get_keywords()
                self["keywords"].keywords = P["extra keywords"].value
        else:
            self.dialog.deactivate(result)
            raise RuntimeError("Don't recognize dialog result '{}'".format(result))

    def help(self):
        """Base class for presenting help, does nothing.

        Subclasses should override this.
        """
        pass

    def initialize_results(self):
        """Initialize the results if empty.

        When the GUI for the step is first created the `results` parameter is empty.
        However the default is to save properties to the database, so they need to
        be put into the `results` parameter.
        """
        if self.node.parameters is None or "results" not in self.node.parameters:
            return

        results = self.node.parameters["results"].value
        if len(results) == 0:
            for key, entry in self.node.metadata["results"].items():
                if "dimensionality" not in entry:
                    continue
                if self.node.calculation is not None and "calculation" in entry:
                    if self.node.calculation not in entry["calculation"]:
                        continue
                if self.node.method is not None and "methods" in entry:
                    if self.node.method not in entry["methods"]:
                        continue

                if "property" in entry:
                    results[key] = {"property": entry["property"]}

    @staticmethod
    def is_expr(value):
        """Return whether the value is an expression or constant.

        Parameters
        ----------
        value : str
            The value to test

        Returns
        -------
        bool
           True for an expression, False otherwise.
        """
        return len(value) > 0 and value[0] == "$"

    def is_inside(self, x, y, halo=0):
        """Return a boolean indicating whether the point x, y is inside
        this node, using halo as a size around the point
        """
        if x < self.x - self.w / 2 - halo:
            return False
        if x > self.x + self.w / 2 + halo:
            return False

        if y < self.y - self.h / 2 - halo:
            return False
        if y > self.y + self.h / 2 + halo:
            return False

        return True

    def move(self, deltax, deltay):
        """Move the node on the canvas.

        Parameters
        ----------
        deltax : int
            The number of pixels to move in the x-direction.
        deltay : int
            The number of pixels to move in the y-direction.
        """
        if self._tmp is None:
            self._tmp = self.connections()

        self.x += deltax
        self.y += deltay

        self.canvas.move(self.tag, deltax, deltay)

        for connection in self._tmp:
            direction, edge = connection
            edge.move()

    def next_anchor(self):
        """Return where the next node should be positioned. The default is
        <gap> below the 's' anchor point.
        """

        return "s"

    def remove_edge(self, edge):
        """Remove a given edge, or all edges if 'all' is given"""

        if isinstance(edge, str) and edge == "all":
            for direction, obj in self.connections():
                self.remove_edge(obj)
        else:
            self.tk_flowchart.graph.remove_edge(
                edge.node1, edge.node2, edge.edge_type, edge.edge_subtype
            )

    def reset_dialog(self, widget=None):
        """Reset the layout of the dialog as needed for the parameters.

        In this base class this does nothing. Override as needed in the
        subclasses derived from this class.
        """
        pass

    def right_click(self, event):
        """Respond to a right-click by posting the popup menu.

        This method provides a popup menu with a **delete** command.

        Subclasses should override or extend this as appropriate! The menu created in
        this base method is accessible in subclasses which should make it easy to
        override.
        """

        if self.popup_menu is not None:
            self.popup_menu.destroy()

        self.popup_menu = tk.Menu(self.canvas, tearoff=0)
        self.popup_menu.add_command(
            label="Delete",
            command=lambda node=self: self.tk_flowchart.remove_node(node),
        )

        if type(self) is seamm.tk_node.TkNode:
            self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def setup_results(self):
        """Layout the results tab of the dialog"""
        if self.node.parameters is None or "results" not in self.node.parameters:
            return

        results = self.node.parameters["results"].value

        # Find what tables are in use.
        tables = set()
        for tmp in results.values():
            if "table" in tmp:
                tables.add(tmp["table"])
        self.node.tables = sorted(tables)

        tables.update(self.node.existing_tables())
        self._tables = sorted(tables)
        del tables

        self.results_widgets = []
        table = self["results"]
        table.clear()
        frame = table.interior()

        row = 0
        for key, entry in self.node.metadata["results"].items():
            if "dimensionality" not in entry:
                continue
            if self.node.calculation is not None and "calculation" in entry:
                if self.node.calculation not in entry["calculation"]:
                    continue
            if self.node.method is not None and "methods" in entry:
                if self.node.method not in entry["methods"]:
                    continue

            widgets = []
            widgets.append(key)

            table.cell(row, 0, entry["description"])

            # Property for DB. default is save
            if "property" in entry:
                var = self.tk_var["property " + key] = {
                    "value": tk.IntVar(),
                    "key": entry["property"],
                }
                if key in results and "property" in results[key]:
                    var["value"].set(1)
                else:
                    var["value"].set(0)

                w = ttk.Checkbutton(frame, variable=var["value"])
                table.cell(row, 1, w)
                widgets.append(w)
            else:
                widgets.append(None)

            # variable
            var = self.tk_var[key] = tk.IntVar()
            var.set(0)
            w = ttk.Checkbutton(frame, variable=var)
            table.cell(row, 2, w)
            widgets.append(w)
            e = ttk.Entry(frame, width=15)
            e.insert(0, key.lower().replace(" ", "_"))
            table.cell(row, 3, e)
            widgets.append(e)

            if key in results:
                if "variable" in results[key]:
                    var.set(1)
                    e.delete(0, tk.END)
                    e.insert(0, results[key]["variable"])

            # table
            w = ttk.Combobox(frame, width=10, values=["", *self._tables])
            table.cell(row, 4, w)
            widgets.append(w)
            w.bind("<<ComboboxSelected>>", self._table_cb)
            w.bind("<Return>", self._table_cb)
            w.bind("<FocusOut>", self._table_cb)
            e = ttk.Entry(frame, width=15)
            e.insert(0, key.lower().replace("_", " "))
            table.cell(row, 5, e)
            widgets.append(e)

            if key in results:
                if "table" in results[key]:
                    w.set(results[key]["table"])
                    e.delete(0, tk.END)
                    e.insert(0, results[key]["column"])

            # And units....
            if "units" in entry and entry["units"] != "":
                units = entry["units"]
                w = ttk.Combobox(frame, width=10)
                widgets.append(w)
                table.cell(row, 6, w)
                w.config(values=[*default_units(units), ""])
                w.set(units)
                if key in results and "units" in results[key]:
                    w.set(results[key]["units"])
            else:
                widgets.append(None)

            self.results_widgets.append(widgets)
            row += 1

    def set_uuid(self):
        """Set the unique id of the node to a new uuid."""
        self.node.set_uuid()

    def to_dict(self):
        """Serialize to a dict"""
        data = {
            "x": self._x,
            "y": self._y,
            "w": self._w,
            "h": self._h,
        }

        return data

    def _table_cb(self, event):
        "Update the list of tables as needed."
        table = event.widget.get()

        if table.strip() == "":
            return

        if table not in self._tables:
            self._tables.append(table)
            self._tables = sorted(self._tables)
            self.node.tables.append(table)
            self.node.tables.sort()

            for (
                key,
                w_property,
                w_check,
                w_variable,
                w_table,
                w_column,
                w_units,
            ) in self.results_widgets:
                if w_table is not None:
                    w_table.configure(values=["", *self._tables])

    def update_flowchart(self, tk_flowchart=None, flowchart=None):
        """Update the nongraphical flowchart. Only used in nodes that contain
        flowcharts"""
        if self.tk_subflowchart is None or self.node.subflowchart is None:
            return

        # Make sure there is nothing in the flowchart
        self.node.subflowchart.clear(all=True)

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in self.tk_subflowchart:
            translate[node] = self.node.subflowchart.add_node(copy.copy(node.node))
            node.update_flowchart()

        # And the edges
        for edge in self.tk_subflowchart.edges():
            attr = {}
            for key in edge:
                if key not in ("node1", "node2", "edge_type", "edge_subtype", "canvas"):
                    attr[key] = edge[key]
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            self.node.subflowchart.add_edge(
                node1, node2, edge.edge_type, edge.edge_subtype, **attr
            )

    def undraw(self):
        """Remove all the visual components from the canvas."""
        self.canvas.delete(self.tag)



# -*- coding: utf-8 -*-

import collections.abc
import logging
import seamm
import pprint

"""A simple graph structure for holding the flowchart. This handles a
directed graph -- all edges have a direction implied -- with zero or
more edges from or to each node.
"""

logger = logging.getLogger(__name__)


class Graph(object):
    """A datastructure for holding a directed graph with multiple (parallel) edges."""

    def __init__(self):
        """Create the graph object"""

        self._node = {}
        self._edge = {}

    def __iter__(self):
        return self._node.values().__iter__()

    def __contains__(self, node):
        return node.uuid in self._node

    def add_node(self, node):
        if node in self:
            raise RuntimeError("node is already in the graph")
        self._node[node.uuid] = node
        return node

    def remove_node(self, node):
        if node not in self:
            raise RuntimeError("node is not in the graph")
        del self._node[node.uuid]

    def clear(self):
        self._node = {}
        self._edge = {}

    def add_edge(
        self, u, v, edge_type=None, edge_subtype=None, edge_class=None, **kwargs
    ):
        if u not in self:
            self.add_node(u)

        if v not in self:
            self.add_node(v)

        key = (u.uuid, v.uuid, edge_type, edge_subtype)
        if edge_class is None:
            self._edge[key] = seamm.Edge(
                self, u, v, edge_type=edge_type, edge_subtype=edge_subtype, **kwargs
            )
        else:
            self._edge[key] = edge_class(
                self, u, v, edge_type=edge_type, edge_subtype=edge_subtype, **kwargs
            )
        return self._edge[key]

    def remove_edge(self, u, v, edge_type=None, edge_subtype=None):
        key = (u.uuid, v.uuid, edge_type, edge_subtype)
        if key not in self._edge:
            raise RuntimeError("edge does not exist!")
        del self._edge[key]

    def edges(self, node=None, direction="both"):
        result = []
        if node is None:
            return self._edge.values()
        else:
            h = node.uuid
            if direction == "both":
                for key in self._edge:
                    h1, h2, edge_type, edge_subtype = key
                    if h1 == h:
                        result.append(("out", self._edge[key]))
                    if h2 == h:
                        result.append(("in", self._edge[key]))
            elif direction == "out":
                for key in self._edge:
                    h1, h2, edge_type, edge_subtype = key
                    if h1 == h:
                        result.append(self._edge[key])
            elif direction == "in":
                for key in self._edge:
                    h1, h2, edge_type, edge_subtype = key
                    if h2 == h:
                        result.append(self._edge[key])
            else:
                return RuntimeError("Don't recognize direction '{}'!".format(direction))

        return result

    def has_edge(self, u, v, edge_type=None, edge_subtype=None):
        key = (u.uuid, v.uuid, edge_type, edge_subtype)
        return key in self._edge


class Edge(collections.abc.MutableMapping):
    def __init__(
        self, graph, node1, node2, edge_type="execution", edge_subtype="next", **kwargs
    ):
        self.graph = graph
        self._data = dict(**kwargs)
        self._data["node1"] = node1
        self._data["node2"] = node2
        self._data["edge_type"] = edge_type
        self._data["edge_subtype"] = edge_subtype

    def __getitem__(self, key):
        """Allow [] access to the dictionary!"""
        return self._data[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._data[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        del self._data[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._data)

    def __len__(self):
        """The len() command"""
        return len(self._data)

    def __repr__(self):
        """The string representation of this object"""
        return repr(self._data)

    def __str__(self):
        """The pretty string representation of this object"""
        return pprint.pformat(self._data)

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        if item in self._data:
            return True
        return False

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self._data == other._data

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self._data.copy()

    @property
    def node1(self):
        return self._data["node1"]

    @property
    def node2(self):
        return self._data["node2"]

    @property
    def edge_type(self):
        return self._data["edge_type"]

    @property
    def edge_subtype(self):
        return self._data["edge_subtype"]


if __name__ == "__main__":

    class Node(object):
        def __init__(self, **kwargs):
            self.data = dict(**kwargs)

        def __str__(self):
            return pprint.pformat(self.data)

    graph = Graph()

    start = Node(title="start")
    node1 = Node(title="node1")
    node2 = Node(title="node2")

    edge1 = graph.add_edge(start, node1)
    print("edge1 = {}".format(edge1))
    edge2 = graph.add_edge(node1, node2)
    print("edge2 = {}".format(edge2))

    print("nodes:")
    for node in graph:
        print(" node = {}".format(node))
        print()

    print("edges:")
    for edge in graph.edges():
        print(edge)
        print()



# -*- coding: utf-8 -*-

"""The GUI for publishing -- flowcharts for the moment."""

import collections.abc
import logging
import tkinter as tk
import tkinter.ttk as ttk

import Pmw

import seamm_util
import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkPublish(collections.abc.MutableMapping):
    def __init__(self, tk_flowchart):
        self.tk_flowchart = tk_flowchart
        self._widgets = {"parent": tk_flowchart.pw}

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets!"""
        return self._widgets[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._widgets[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self._widgets:
            self._widgets[key].destroy()
        del self._widgets[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._widgets)

    def __len__(self):
        """The len() command"""
        return len(self._widgets)

    def create_dialog(self):
        """Create the dialog for publishing."""
        if "dialog" in self:
            return

        toplevel = self["parent"].winfo_toplevel()
        self["dialog"] = d = Pmw.Dialog(
            toplevel,
            buttons=("Publish", "Cancel"),
            master=toplevel,
            title="Publish the flowchart",
        )
        d.withdraw()

        frame = self["frame"] = d.interior()
        # P = self.parameters
        # for item in ("what", "destination"):
        #     self[item] = P[item].widget(frame)

        w = self["what"] = sw.LabeledCombobox(
            frame,
            labeltext="Publish the",
            values=("flowchart",),
            state="readonly",
        )
        w.set("flowchart")

        w = self["destination"] = sw.LabeledCombobox(
            frame,
            labeltext="to",
            values=("Zenodo", "Zenodo sandbox"),
            state="readonly",
        )
        w.set("Zenodo")

        for item in ("what",):
            self[item].bind("<<ComboboxSelected>>", self.reset_dialog)
            self[item].bind("<Return>", self.reset_dialog)
            self[item].bind("<FocusOut>", self.reset_dialog)

        # Need the start node for editing the flowchart metadata
        start_node = self.tk_flowchart.get_node("1")
        self["flowchart metadata"] = start_node.create_frame(frame)

        self["doi"] = ttk.Label(frame, text="")

        self.reset_dialog()

    def edit(self):
        """Present a dialog for editing the parameters."""
        # Create the dialog if it doesn't exist
        self.create_dialog()

        # update the flowchart metadata
        start_node = self.tk_flowchart.get_node("1")
        start_node.update_widgets()

        self.reset_dialog()

        # And put it on-screen, the first time centered.
        result = self["dialog"].activate(geometry="centerscreenfirst")

        if result != "Publish":
            # Reset everything!
            start_node.cleanup_authors()
        else:
            # Handle the metadata first
            start_node.capture_metadata()
            start_node.cleanup_authors()

            # Capture the parameters from the widgets
            what = self["what"].get()
            destination = self["destination"].get()

            if what == "flowchart":
                if destination == "Zenodo":
                    doi = self.publish_flowchart_to_zenodo()
                elif destination == "Zenodo sandbox":
                    doi = self.publish_flowchart_to_zenodo(sandbox=True)
                else:
                    raise RuntimeError(f"cannot handle destination '{destination}'")
                tk.messagebox.showinfo(
                    master=self["parent"],
                    title="Published",
                    message=f"Published the {what} to {destination}\nDOI = {doi}",
                )
            else:
                raise RuntimeError(f"cannot handle publishing '{what}'")

    def reset_dialog(self):
        """Layout the widgets in the dialog according to the parameters."""

        what = self["what"].get()

        # Temporary
        if what != "flowchart":
            what = "flowchart"
            self["what"].set(what)

        # Remove all the widgets
        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # and put them back in as needed.
        row = 0
        self["what"].grid(row=row, column=0, sticky=tk.EW)
        row += 1
        self["destination"].grid(row=row, column=0, sticky=tk.EW)
        row += 1

        if what == "flowchart":
            self["flowchart metadata"].grid(row=row, column=0, sticky=tk.EW)
            row += 1

            self["doi"].grid(row=row, column=0, sticky=tk.W)
            row += 1
            metadata = self.tk_flowchart.flowchart.metadata
            if "doi" in metadata:
                self["doi"]["text"] = metadata["doi"]
            else:
                self["doi"]["text"] = "not published"

    def publish_flowchart_to_zenodo(self, sandbox=False):
        """Publish the flowchart to Zenodo.

        Parameters
        ----------
        sandbox : bool = False
            If true, publish to the Zenodo sandbox.

        Returns
        -------
        str
            The DOI.
        """
        flowchart = self.tk_flowchart.flowchart

        # Get check if the flowchart is fully released
        if sandbox:
            if flowchart.is_development:
                print("Publishing a dirty version to the Zenodo sandbox.")
                logger.warning("Publishing a dirty version to the Zenodo sandbox.")
        else:
            if flowchart.is_development:
                return None

        zenodo = seamm_util.Zenodo(use_sandbox=sandbox)

        # If this flowchart has an existing DOI, create a new version.
        metadata = flowchart.metadata
        if "doi" in metadata:
            _id = metadata["doi"].split("/zenodo.")[1]
            record = zenodo.add_version(_id)
            for filename in record.files():
                record.remove_file(filename)
        else:
            record = zenodo.create_record()

        # Set up the metadata for Zenodo.
        record.upload_type = "other"
        record.creators = metadata["creators"]
        record.title = metadata["title"]
        record.description = metadata["description"]
        record.keywords = ["seamm-flowchart", *metadata["keywords"]]
        if "grants" in metadata:
            record.grants = [{"id": g} for g in metadata["grants"]]

        # Update the metadata at Zenodo
        record.update_metadata()

        # Update the metadata in the flowchart
        metadata["doi"] = record.doi
        conceptdoi = record.conceptdoi
        if conceptdoi is not None:
            metadata["conceptdoi"] = conceptdoi

        # Now we can add the flowchart to Zenodo with the correct DOI, etc.
        record.add_file("flowchart.flow", contents=flowchart.to_text())

        # And, finally, can publish!
        record.publish()

        return record.doi



# -*- coding: utf-8 -*-

"""A node to join the flow in a flowchart"""

import logging
import seamm

logger = logging.getLogger(__name__)


class TkJoin(seamm.TkNode):
    """The Tk-based graphical representation of a joining node"""

    anchor_points = {
        "n": (0, -0.5),
        "s": (0, 0.5),
        "e": (0.5, 0.0),
        "w": (-0.5, 0.0),
    }

    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=120, y=20, w=30, h=30
    ):
        """Initialize a node

        Keyword arguments:
        """
        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def draw(self):
        """Draw the node on the given canvas, making it visible"""
        # Remove any graphics items
        # self.undraw()

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y - self.h / 2
        y1 = y0 + self.h
        self.border = self.canvas.create_oval(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, "type=outline"],
            fill="white",
        )

        for direction, edge in self.connections():
            edge.move()

    def right_click(self, event):
        """
        Handles the right click event on the node.

        Parameters
        ----------
        event : Tk Event

        Returns
        -------
        None

        See Also
        --------
        TkGaussian.edit
        """

        super().right_click(event)
        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)



# -*- coding: utf-8 -*-

"""Run a SEAMM flowchart.

SEAMM flowcharts have a 'magic' line, so that they can be executed directly.
Or, run_flowchart can be invoked with the name of flowchart.
"""

import configparser
from datetime import datetime, timedelta, timezone
import fasteners
import json
import locale
import logging
import os
import os.path
from pathlib import Path
import platform
import re
import shutil
import sqlite3
import sys
import textwrap
import time
import traceback
import uuid

import cpuinfo

import seamm
import seamm_datastore

import seamm_util

printer = seamm_util.printing.getPrinter()

logging.basicConfig(level="WARNING")
logger = logging.getLogger(__name__)
variables = seamm.Variables()
header_line = "!MolSSI job_data 1.0\n"


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def open_datastore(root, datastore):
    """Open the database via the datastore"""
    # Get the user information for the datastore
    path = Path("~/.seammrc").expanduser()
    if not path.exists:
        raise RuntimeError(
            "You need a '~/.seammrc' file to run jobs from the commandline. "
            "See the documentation for more details."
        )

    config = configparser.ConfigParser()
    config.read(path)

    user = None
    password = None

    if "dev" in root.lower():
        sections = ["dev"]
    else:
        sections = ["localhost"]
    sections.append(platform.node())
    for section in sections:
        section = "Dashboard: " + section
        if section in config:
            if user is None and "user" in config[section]:
                user = config[section]["user"]
            if password is None and "password" in config[section]:
                password = config[section]["password"]

    if user is None or password is None:
        raise RuntimeError(
            "You need credentials in '~/.seammrc' file to run jobs from the "
            "commandline. See the documentation for more details."
        )

    # Add to the database
    db_path = Path(datastore).expanduser().resolve() / "seamm.db"
    db_uri = "sqlite:///" + str(db_path)
    db = seamm_datastore.connect(
        database_uri=db_uri,
        datastore_location=datastore,
        username=user,
        password=password,
    )

    return db


def run_from_jobserver():
    """Helper routine to run from the JobServer.

    Gets the arguments from the command line.
    """
    job_id = sys.argv[1]
    wdir = sys.argv[2]
    db_path = sys.argv[3]
    cmdline = sys.argv[4:]

    with cd(wdir):
        try:
            run(
                job_id=job_id,
                wdir=wdir,
                db_path=db_path,
                in_jobserver=True,
                cmdline=cmdline,
            )
        except Exception as e:
            path = Path("job_data.json")
            if path.exists():
                with path.open("r") as fd:
                    fd.readline()
                    data = json.load(fd)
            else:
                data = {}

            data["state"] = "error"
            data["error type"] = type(e).__name__
            data["error message"] = traceback.format_exc()
            with path.open("w") as fd:
                fd.write("!MolSSI job_data 1.0")
                json.dump(data, fd, indent=3, sort_keys=True)
                fd.write("\n")
            raise


def run(
    job_id=None,
    wdir=None,
    db_path=None,
    setup_logging=True,
    in_jobserver=False,
    cmdline=None,
):
    """The standalone flowchart app"""
    global print

    if not in_jobserver and len(sys.argv) > 1:
        if sys.argv[1] == "--help" or sys.argv[1] == "-h":
            # Running run_flowchart by hand ...
            print("usage: run_flowchart <flowchart> [options]")
            print("")
            print("usually it is simpler to execute the flowchart file itself")
            exit()

        # Slice off 'run_flowchart' from the arguments, leaving the
        # flowchart as the thing being run.
        sys.argv = sys.argv[1:]

        filename = sys.argv[0]
    else:
        if wdir is None:
            filename = "flowchart.flow"
        else:
            filename = os.path.join(wdir, "flowchart.flow")

    if cmdline is None:
        cmdline = sys.argv[1:]

    # Set up the argument parser for this node.
    parser = seamm_util.seamm_parser()

    parser.epilog = textwrap.dedent(
        """
        The plug-ins in this flowchart are listed above.
        Options, if any, for plug-ins are placed after
        the name of the plug-in, e.g.:

           test.flow lammps-step --log-level DEBUG --np 4

        To get help for a plug-in, use --help or -h after the
        plug-in name. E.g.

           test.flow lammps-step --help
        """
    )
    parser.usage = "%(prog)s [options] plug-in [options] plug-in [options] ..."

    # Now we need to get the flowchart so that we can set up all the
    # parsers for the steps in order to provide appropriate help.
    if not os.path.exists(filename):
        raise FileNotFoundError(f"The flowchart '{filename}' does not exist.")

    logger.info(f"    reading in flowchart '{filename}'")
    flowchart = seamm.Flowchart(parser_name="SEAMM")
    flowchart.read(filename)
    logger.info("   finished reading the flowchart")

    # Now traverse the flowchart, setting up the ids and parsers
    flowchart.set_ids()
    flowchart.create_parsers()

    # And handle the command-line arguments and ini file options.
    parser.parse_args(cmdline)
    logger.info("Parsed the command-line arguments")
    options = parser.get_options("SEAMM")

    # Whether to just run as-is, without getting a job_id, using the
    # datastore, etc.
    standalone = options["standalone"] or options["projects"] is None

    # Setup the logging
    if setup_logging:
        if "log_level" in options:
            logging.basicConfig(level=options["log_level"])

        # Set the log level for the plug-ins
        flowchart.set_log_level(parser.get_options())

    # Create the working directory where files, output, etc. go.
    # At the moment this is datastore/job_id

    if standalone:
        print("Running in standalone mode.")
        if wdir is None:
            wdir = os.getcwd()
    else:
        datastore = os.path.expanduser(options["datastore"])

        if job_id is None:
            if options["job_id_file"] is None:
                job_id_file = os.path.join(datastore, "job.id")
            else:
                job_id_file = options["job_id_file"]

            # Get the job_id from the file, creating the file if necessary
            job_id = get_job_id(job_id_file)
        if options["projects"] is None:
            projects = ["default"]
        else:
            projects = options["projects"]
        if wdir is None:
            # And put it all together
            wdir = os.path.abspath(
                os.path.join(
                    datastore, "projects", projects[0], "Job_{:06d}".format(job_id)
                )
            )

            if os.path.exists(wdir):
                if options["force"]:
                    shutil.rmtree(wdir)
                else:
                    msg = "Directory '{}' exists, use --force to overwrite".format(wdir)

                    logging.critical(msg)
                    sys.exit(msg)

            os.makedirs(wdir, exist_ok=False)

    logging.info("The working directory is '{}'".format(wdir))

    # Set up the root printer, and add handlers to print to the file
    # 'job.out' in the working directory and to stdout, as requested
    # in the options. Since all printers are children of the root
    # printer, all output at the right levels will flow here

    # Set up our formatter
    formatter = logging.Formatter(fmt="{message:s}", style="{")

    # A handler for stdout
    if not in_jobserver:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(seamm_util.printing.NORMAL)
        console_handler.setFormatter(formatter)
        printer.addHandler(console_handler)

    # A handler for the file
    # Frist remove the job.out file if it exists so we start freash
    job_file = Path(wdir) / "job.out"
    job_file.unlink(missing_ok=True)

    file_handler = logging.FileHandler(os.path.join(wdir, "job.out"))
    file_handler.setLevel(seamm_util.printing.NORMAL)
    file_handler.setFormatter(formatter)
    printer.addHandler(file_handler)

    # And ... finally ... run!
    printer.job(datetime.now().strftime("%A %Y.%m.%d %H:%M:%S %Z"))
    printer.job("Running in directory '{}'".format(wdir))

    flowchart_path = Path(wdir).resolve() / "flowchart.flow"
    path = Path(sys.argv[0]).resolve()

    # copy the flowchart to the root directory if it is not there already
    if not in_jobserver and flowchart_path.exists() and path != flowchart_path:
        shutil.copy2(path, flowchart_path)

    # Make executable if it isn't
    permissions = flowchart_path.stat().st_mode
    if permissions & 0o100 == 0:
        flowchart_path.chmod(permissions | 0o110)

    logger.info(f"    reading in flowchart '{flowchart_path}' -- 2")
    flowchart = seamm.Flowchart(directory=wdir)
    flowchart.read(flowchart_path)
    logger.info("   finished reading the flowchart -- 2")

    # Change to the working directory and run the flowchart
    with cd(wdir):
        # Set up the initial metadata for the job.
        time_now = datetime.now(timezone.utc).isoformat()
        if in_jobserver:
            with open("job_data.json", "r") as fd:
                fd.readline()
                data = json.load(fd)
        else:
            if options["title"] != "":
                title = options["title"]
            else:
                title = flowchart.metadata["title"]
            if title == "":
                "untitled job"
            if "description" in options:
                description = options["description"]
            else:
                description = flowchart.metadata["description"]
            if description == "":
                description = "Run from the command-line."
            data = {
                "data_version": "1.0",
                "title": title,
                "working directory": wdir,
                "submitted time": time_now,
            }
        data.update(
            {
                "command line": cmdline,
                "flowchart_digest": flowchart.digest(),
                "flowchart_digest_strict": flowchart.digest(strict=True),
                "start time": time_now,
                "state": "started",
                "uuid": uuid.uuid4().hex,
                "~cpuinfo": cpuinfo.get_cpu_info(),
            }
        )
        if not in_jobserver and not standalone:
            if "projects" not in data:
                data["projects"] = projects
            data["datastore"] = datastore
            data["job id"] = job_id

            db = open_datastore(options["root"], datastore)

            pid = os.getpid()
            current_time = datetime.now(timezone.utc)
            with seamm_datastore.session_scope(db.Session) as session:
                job = db.Job.create(
                    job_id,
                    str(flowchart_path),
                    project_names=data["projects"],
                    path=wdir,
                    title=title,
                    description=description,
                    submitted=current_time,
                    started=current_time,
                    parameters={"cmdline": [], "pid": pid},
                    status="started",
                )
                session.add(job)

            del db

        # Output the initial metadata for the job.
        with open("job_data.json", "w") as fd:
            fd.write(header_line)
            json.dump(data, fd, indent=3, sort_keys=True)
            fd.write("\n")

        t0 = time.time()
        pt0 = time.process_time()

        # And run the flowchart
        logger.info("Executing the flowchart")
        try:
            exec = seamm.ExecFlowchart(flowchart)
            exec.run(root=wdir)
            data["state"] = "finished"
        except Exception as e:
            data["state"] = "error"
            data["error type"] = type(e).__name__
            data["error message"] = traceback.format_exc()
            printer.job(traceback.format_exc())
        finally:
            # Wrap things up
            t1 = time.time()
            pt1 = time.process_time()
            data["end time"] = datetime.now(timezone.utc).isoformat()
            t = t1 - t0
            pt = pt1 - pt0
            data["elapsed time"] = t
            data["process time"] = pt

            with open("job_data.json", "w") as fd:
                fd.write(header_line)
                json.dump(data, fd, indent=3, sort_keys=True)
                fd.write("\n")

            printer.job(f"\nProcess time: {timedelta(seconds=pt)} ({pt:.3f} s)")
            printer.job(f"Elapsed time: {timedelta(seconds=t)} ({t:.3f} s)")

            if in_jobserver:
                datastore = os.path.expanduser(options["datastore"])
                try:
                    current_time = datetime.now(timezone.utc)
                    # Open the database directly, relying on file permissions
                    if db_path is None:
                        db_path = Path(datastore).expanduser().resolve() / "seamm.db"
                    db = sqlite3.connect(db_path)
                    cursor = db.cursor()
                    cursor.execute(
                        "UPDATE jobs"
                        "   SET status = ?, finished = ?,"
                        "       parameters=json_remove(jobs.parameters, '$.pid')"
                        " WHERE id = ?",
                        (data["state"], current_time, job_id),
                    )
                    db.commit()
                    db.close()
                except Exception as e:
                    printer.job(e)
            elif not standalone:
                # Let the datastore know that the job finished.
                # current_time = datetime.now(timezone.utc)

                # At the moment update takes weird numbers!
                now = datetime.now().astimezone()
                dt = now.utcoffset().total_seconds()
                current_time = (time.time() - dt) * 1000

                # Add to the database
                # N.B. Don't know how to remove 'pid' from the JSON properties column
                db = open_datastore(options["root"], datastore)
                with seamm_datastore.session_scope(db.Session) as session:
                    job = db.Job.update(
                        job_id, finished=current_time, status=data["state"]
                    )
                del db
        printer.job(datetime.now().strftime("%A %Y.%m.%d %H:%M:%S %Z"))


def get_job_id(filename):
    """Get the next job id from the given file.

    This uses the fasteners module to provide locking so that
    only one job at a time can access the file, so that the job
    ids are unique and monotonically increasing.
    """

    filename = os.path.expanduser(filename)

    lock_file = filename + ".lock"
    lock = fasteners.InterProcessLock(lock_file)
    locked = lock.acquire(blocking=True, timeout=5)

    if locked:
        if not os.path.isfile(filename):
            job_id = 1
            with open(filename, "w") as fd:
                fd.write("!MolSSI job_id 1.0\n")
                fd.write("1\n")
            lock.release()
        else:
            with open(filename, "r+") as fd:
                line = fd.readline()
                pos = fd.tell()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is empty".format(filename))
                line = line.strip()
                match = re.fullmatch(r"!MolSSI job_id ([0-9]+(?:\.[0-9]+)*)", line)
                if match is None:
                    lock.release()
                    raise RuntimeError(
                        "The job_id file has an incorrect header: {}".format(line)
                    )
                line = fd.readline()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is truncated".format(filename))
                try:
                    job_id = int(line)
                except TypeError:
                    raise TypeError(
                        "The job_id in file '{}' is not an integer: {}".format(
                            filename, line
                        )
                    )
                job_id += 1
                fd.seek(pos)
                fd.write("{:d}\n".format(job_id))
    else:
        raise RuntimeError("Could not lock the job_id file '{}'".format(filename))

    return job_id


if __name__ == "__main__":
    locale.setlocale(locale.LC_ALL, "")
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The interface for submitting SEAMM jobs.

A job in SEAMM is composed of a flowchart and any other files that the
flowchart requires. This module provides the JobHandler class, which
provides a use interface and the machinery to gather the necessary
files and submit the job to a dashboard.
"""

import configparser
import logging
from pathlib import Path
import pkg_resources

from .seammrc import SEAMMrc
import seamm_dashboard_client
import seamm_util

logger = logging.getLogger(__name__)


def safe_filename(filename):
    if filename[0] == "~":
        path = Path(filename).expanduser()
    else:
        path = Path(filename)
    if path.anchor == "":
        result = "_".join(path.parts)
    else:
        result = "_".join(path.parts[1:])
    return "job:data/" + result


class DashboardHandler(object):
    def __init__(self, user_agent=None):
        """Setup the handler for Dashboards.

        Parameters
        ----------
        """
        self.config = configparser.ConfigParser()

        if user_agent is None:
            import seamm

            self.user_agent = f"SEAMM/{seamm.__version__}"
        else:
            self.user_agent = user_agent

        self._credentials = SEAMMrc()
        self._current_dashboard = None
        self.resource_path = Path(pkg_resources.resource_filename(__name__, "data/"))

        # Get the location of the dashboards configuration file
        parser = seamm_util.seamm_parser()
        options = parser.get_options()
        if "dashboards" in options:
            self.configfile = Path(options["dashboards"]).expanduser()
        else:
            self.configfile = Path.home() / "SEAMM" / "dashboards.ini"

        # Create the file if it doesn't exist
        if not self.configfile.exists():
            self.configfile.parent.mkdir(parents=True, exist_ok=True)
            path = self.resource_path / "dashboards.ini"
            text = path.read_text()
            self.configfile.write_text(text)

        # Read in the dashboard information if present
        self.get_configuration()
        self.current_dashboard = self.config.get(
            "GENERAL", "current_dashboard", fallback=None
        )
        self.timeout = self.config.get("GENERAL", "timeout", fallback=1)

    @property
    def credentials(self):
        """The data Dashboard from ~/.seammrc."""
        return self._credentials

    @property
    def current_dashboard(self):
        "The currently selected dashboard"
        if self._current_dashboard is None:
            dashboards = self.dashboards

            dashboard = self.config.get(
                "GENERAL", "current_dashboard", fallback=dashboards[0]
            )
            if dashboard not in dashboards:
                dashboard = dashboards[0]

            self._current_dashboard = self.get_dashboard(dashboard)
            self.save_configuration()
        return self._current_dashboard

    @current_dashboard.setter
    def current_dashboard(self, dashboard):
        if isinstance(dashboard, str):
            self._current_dashboard = self.get_dashboard(dashboard)
        else:
            if dashboard.name in self.dashboards:
                self._current_dashboard = dashboard
            else:
                raise ValueError(f"Dashboard {dashboard.name} does not exist!")

    @property
    def dashboards(self):
        """The list of dashboards."""
        result = []
        for section in self.config.sections():
            if section not in ("GENERAL", self.config.default_section):
                result.append(section)
        return sorted(result)

    def add_dashboard(self, name, url, protocol):
        "Add a new dashboard to the config file"
        self.config[name] = {"url": url, "protocol": protocol}
        self.save_configuration()

    def get_all_status(self):
        """Get the status of all the dashboards.

        Parameters
        ----------
        """
        result = []
        for dashboard in self.dashboards:
            status = self.get_dashboard(dashboard).status()
            result.append((dashboard, status))

        return result

    def get_configuration(self):
        """Get the list of dashboards from the config file."""
        # The path to the configfile
        if self.configfile.exists():
            self.config.read(self.configfile)
        else:
            self.config.clear()

    def get_credentials(self, dashboard, ask=None):
        """The user for the dashboard.

        Parameters
        ----------
        dashboard : str
            The name of the dashboard to use.

        Returns
        -------
        str, str
            The user name and password
        ask : function
            A function or method to call to get the user name and passwd.
        """
        self.credentials.re_read()
        user = None
        password = None
        section = f"Dashboard: {dashboard}"
        if section not in self.credentials:
            self.credentials.add_section(section)

        user = self.credentials.get(section, "user", fallback=None)
        password = self.credentials.get(section, "password", fallback=None)

        if ask is not None and (user is None or password is None):
            user, password = ask(dashboard, user=user, password=password)
            if user is not None and password is not None:
                self.credentials.set(section, "user", user)
                self.credentials.set(section, "password", password)
        return user, password

    def get_dashboard(self, name):
        """Get the given dashboard object.

        Parameters
        ----------
        name : str
            The name of the Dashboard

        Returns
        -------
        seamm_dashboard_client.Dashboard :
            The Dashboard client object.
        """
        user, passwd = self.get_credentials(name)
        url = self.config[name]["url"]

        return seamm_dashboard_client.Dashboard(
            name, url, username=user, password=passwd, user_agent=self.user_agent
        )

    def rename_dashboard(self, old, new):
        "Rename a dashboard from 'old' to 'new'."
        tmp = {}
        for key, value in self.config[old].items():
            tmp[key] = value
        self.config.remove_section(old)
        self.config[new] = tmp

    def save_configuration(self):
        """Save the list of dashboards to disk."""
        # Make sure the directory exists
        self.configfile.parent.mkdir(exist_ok=True)

        # Update the current dashboard
        if self.current_dashboard is not None:
            if "GENERAL" not in self.config:
                self.config["GENERAL"] = {}
            defaults = self.config["GENERAL"]
            defaults["current_dashboard"] = self.current_dashboard.name

        with self.configfile.open("w") as fd:
            self.config.write(fd)

    def update(self, dashboard):
        "Update the dashboard with that given."
        self.config[dashboard.name] = {
            "url": dashboard.url,
            "protocol": "http",
        }
        self.save_configuration()
